# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


# TODO: make generic (use Switch module, different switch models, bistable, etc.)
class Relay(Module):
    switch_a_nc: F.Electrical
    switch_a_common: F.Electrical
    switch_a_no: F.Electrical
    switch_b_no: F.Electrical
    switch_b_common: F.Electrical
    switch_b_nc: F.Electrical
    coil_power: F.ElectricPower

    coil_max_voltage = L.p_field(units=P.V)
    coil_max_current = L.p_field(units=P.A)
    coil_resistance = L.p_field(units=P.ohm)
    contact_max_switching_voltage = L.p_field(units=P.V)
    contact_max_switching_current = L.p_field(units=P.A)
    contact_max_current = L.p_field(units=P.A)

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.K
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import datetime
import logging
import re
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Self

from more_itertools import first
from natsort import natsorted

import faebryk.library._F as F
from atopile.config import config as Gcfg
from faebryk.libs.codegen.atocodegen import AtoCodeGen
from faebryk.libs.codegen.atocodeparse import AtoCodeParse
from faebryk.libs.codegen.pycodegen import sanitize_name
from faebryk.libs.kicad.fileformats_common import ChecksumMismatch, PropertyNotSet
from faebryk.libs.kicad.fileformats_latest import (
    C_kicad_footprint_file,
    C_kicad_model_file,
)
from faebryk.libs.kicad.fileformats_sch import C_kicad_sym_file
from faebryk.libs.picker.picker import PickedPart
from faebryk.libs.util import compare_dataclasses, starts_or_ends_replace

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class AtoPart:
    @dataclass(kw_only=True)
    class AutoGenerated:
        system: str = Path(__file__).stem
        date: datetime.datetime = field(
            default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
        )
        source: str | None
        checksum: str | None = None

    identifier: str
    path: Path
    mfn: tuple[str, str]
    fp: C_kicad_footprint_file
    symbol: C_kicad_sym_file
    model: C_kicad_model_file | None
    auto_generated: AutoGenerated | None
    datasheet: str | None = None
    docstring: str = ""
    pick_part: PickedPart | None = None

    @property
    def fp_path(self) -> Path:
        return self.path / f"{self.fp.footprint.base_name}.kicad_mod"

    @property
    def sym_path(self) -> Path:
        return (
            self.path
            / f"{first(self.symbol.kicad_symbol_lib.symbols.values()).name}.kicad_sym"
        )

    @property
    def ato_path(self) -> Path:
        return self.path / (self.path.name + ".ato")

    @property
    def model_path(self) -> Path:
        if not self.model:
            raise ValueError("Model is not set")
        return self.path / self.model.filename

    def __post_init__(self):
        self.fp = deepcopy(self.fp)
        self.fp.footprint.name = f"{self.identifier}:{self.fp.footprint.base_name}"

        self.symbol = deepcopy(self.symbol)

        if self.model:
            rel_path = Gcfg.project.get_relative_to_kicad_project(self.model_path)

            if len(self.fp.footprint.models) != 1:
                raise NotImplementedError(
                    "Multiple models are not supported yet. "
                    "Please report this issue to the maintainers."
                )
            self.fp.footprint.models[0].path = rel_path

    def compare(self, other: Self) -> dict:
        return compare_dataclasses(
            self,
            other,
            skip_keys=(
                "path",
                "uuid",
                "date",
                "checksum",
            ),
        )

    def _dump_pins(self, build: AtoCodeGen.ComponentFile):
        symbol = first(self.symbol.kicad_symbol_lib.symbols.values())

        build.add_comments("pins", use_spacer=True)

        def _sanitize(pin_name: str) -> str | None:
            # number only == pin_number, not name
            if re.match(r"^[0-9]+$", pin_name):
                return None

            # handle negation
            pin_name = starts_or_ends_replace(pin_name, ("~", "#"), prefix="n")
            # handle diff pair
            pin_name = starts_or_ends_replace(pin_name, ("+",), suffix="pos")
            pin_name = starts_or_ends_replace(pin_name, ("-", "–"), suffix="neg")

            return sanitize_name(pin_name, warn_prefix=f"{self.identifier}")

        # Collect and sort pins first
        unsorted_pins = []
        for unit in symbol.symbols.values():
            for pin in unit.pins:
                pin_num = pin.number.number
                pin_name = pin.name.name
                if pin_name.upper() == "NC":
                    continue
                pin_name = _sanitize(pin_name)
                unsorted_pins.append((pin_name, pin_num))

        # Sort pins by name using natsort
        sorted_pins = natsorted(unsorted_pins, key=lambda x: x[0])

        defined_signals = set()

        # Process sorted pins
        for pin_name, pin_num in sorted_pins:
            if pin_name is None:
                build.add_stmt(AtoCodeGen.PinDeclaration(pin_num))
            else:
                build.add_stmt(
                    AtoCodeGen.Connect(
                        left=AtoCodeGen.Connect.Connectable(
                            pin_name,
                            declare="signal"
                            if pin_name not in defined_signals
                            else None,
                        ),
                        right=AtoCodeGen.Connect.Connectable(pin_num, declare="pin"),
                    )
                )
                defined_signals.add(pin_name)

    def dump(self):
        assert self.auto_generated is not None

        self.path.mkdir(parents=True, exist_ok=True)

        # refresh checksums
        self.fp.footprint.set_checksum()
        symbol = first(self.symbol.kicad_symbol_lib.symbols.values())
        symbol.set_checksum()

        self.fp.dumps(self.fp_path)
        self.symbol.dumps(self.sym_path)
        if self.model:
            self.model.dumps(self.model_path)

        ato_builder = AtoCodeGen.ComponentFile(
            f"{self.identifier}_package", docstring=self.docstring
        )
        ato_builder.add_comments(
            "This trait marks this file as auto-generated",
            "If you want to manually change it, remove the trait",
        )
        ato_builder.add_trait(
            "is_auto_generated",
            system=self.auto_generated.system,
            source=self.auto_generated.source,
            date=self.auto_generated.date.isoformat(),
            checksum=F.is_auto_generated.CHECKSUM_PLACEHOLDER,
        )

        ato_builder.add_stmt(AtoCodeGen.Spacer())

        ato_builder.add_trait(
            "is_atomic_part",
            manufacturer=self.mfn[0],
            partnumber=self.mfn[1],
            footprint=self.fp_path.name,
            symbol=self.sym_path.name,
            model=self.model_path.name if self.model else None,
        )

        if self.pick_part:
            ato_builder.add_trait(
                "has_part_picked",
                "by_supplier",
                supplier_id=self.pick_part.supplier.supplier_id,
                supplier_partno=self.pick_part.supplier_partno,
                manufacturer=self.pick_part.manufacturer,
                partno=self.pick_part.partno,
            )

        ato_builder.add_trait(
            "has_designator_prefix", prefix=symbol.propertys["Reference"].value
        )

        if self.datasheet:
            ato_builder.add_trait("has_datasheet_defined", datasheet=self.datasheet)

        self._dump_pins(ato_builder)

        ato = ato_builder.dump()

        ato = F.is_auto_generated.set_checksum(ato)

        self.ato_path.write_text(ato, encoding="utf-8")

    def verify_checksum(self, ato_file_contents: str):
        from faebryk.library.is_auto_generated import _FileManuallyModified

        if not self.auto_generated:
            return
        assert self.auto_generated.checksum is not None

        # check ato
        F.is_auto_generated.verify(self.auto_generated.checksum, ato_file_contents)

        fp = self.fp.footprint
        symbol = first(self.symbol.kicad_symbol_lib.symbols.values())

        # check fp & symbol
        for obj, t_name in ((fp, "Footprint"), (symbol, "Symbol")):
            try:
                obj.verify_checksum()
            except PropertyNotSet:
                raise _FileManuallyModified(
                    f"{t_name} has no checksum."
                    "But part is auto-generated. This is not allowed."
                )
            except ChecksumMismatch:
                raise _FileManuallyModified(
                    f"{t_name} has a checksum mismatch. "
                    "But part is auto-generated. This is not allowed. "
                )

        # TODO verify model

    @classmethod
    def load(cls, path: Path) -> Self:
        """
        Looks for ato manifest in the given directory.
        If found, loads the manifest and returns an AtoPart object.
        """

        # TODO consider using Bob
        ato = AtoCodeParse.ComponentFile(path / (path.name + ".ato"))

        atomic_trait = ato.get_trait(F.is_atomic_part)
        auto_generated_trait = ato.try_get_trait(F.is_auto_generated)
        datasheet = (
            t.get_datasheet()
            if (t := ato.try_get_trait(F.has_datasheet_defined))
            else None
        )

        pick_part = (
            t.get_part() if (t := ato.try_get_trait(F.has_part_picked)) else None
        )

        docstring = ato.parse_docstring()

        fp = C_kicad_footprint_file.loads(path / atomic_trait._footprint)
        symbol = C_kicad_sym_file.loads(path / atomic_trait._symbol)
        model = (
            C_kicad_model_file.loads(path / atomic_trait._model)
            if atomic_trait._model
            else None
        )
        mfn_pn = atomic_trait._manufacturer, atomic_trait._partnumber

        auto_generated = None
        if auto_generated_trait is not None:
            assert auto_generated_trait._system is not None
            assert auto_generated_trait._date is not None

            auto_generated = AtoPart.AutoGenerated(
                system=auto_generated_trait._system,
                date=datetime.datetime.fromisoformat(auto_generated_trait._date),
                source=auto_generated_trait._source,
                checksum=auto_generated_trait._checksum,
            )

        out = cls(
            identifier=path.name,
            path=path,
            mfn=mfn_pn,
            fp=fp,
            symbol=symbol,
            model=model,
            auto_generated=auto_generated,
            datasheet=datasheet,
            docstring=docstring,
            pick_part=pick_part,
        )

        out.verify_checksum(ato.ato)

        return out



# This file is part of the atopile project
# SPDX-License-Identifier: MIT

"""
This file contains an atopile sample.
"""

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.brightness import TypicalLuminousIntensity
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class App(Module):
    led: F.PoweredLED
    battery: F.Battery

    def __preinit__(self) -> None:
        self.led.power.connect(self.battery.power)

        # Parametrize
        self.led.led.color.constrain_subset(F.LED.Color.YELLOW)
        self.led.led.brightness.constrain_subset(
            TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
        )

        self.battery.add(F.has_designator_prefix("B"))
        self.battery.add(F.has_descriptive_properties_defined({"LCSC": "C5239862"}))
        self.battery.voltage.alias_is(3 * P.V)
        self.battery.add(
            F.has_explicit_part.by_supplier(
                "C5239862",
                pinmap={
                    "1": self.battery.power.lv,
                    "2": self.battery.power.hv,
                },
            )
        )

        self.led.add(F.has_designator_prefix("LED"))
        self.led.add(
            F.has_explicit_part.by_supplier(
                "C72038",
                pinmap={
                    "1": self.led.led.cathode,
                    "2": self.led.led.anode,
                },
            )
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.exporters.pcb.layout.heuristic_decoupling import Params, place_next_to
from faebryk.exporters.pcb.layout.layout import Layout
from faebryk.libs.util import not_none

logger = logging.getLogger(__name__)


class LayoutHeuristicElectricalClosenessPullResistors(Layout):
    Parameters = Params

    def __init__(self, params: Params | None = None):
        super().__init__()
        self._params = params or Params()

    def apply(self, *node: Node):
        # Remove nodes that have a position defined
        node = tuple(
            n
            for n in node
            if not n.has_trait(F.has_pcb_position) and n.has_trait(F.has_footprint)
        )

        for n in node:
            assert isinstance(n, F.Resistor)
            logic = not_none(n.get_parent_of_type(F.ElectricLogic))

            place_next_to(logic.line, n, route=True, params=self._params)

    @staticmethod
    def find_module_candidates(node: Node):
        return Module.get_children_modules(
            node,
            direct_only=False,
            types=F.Resistor,
            f_filter=lambda c: c.get_parent_of_type(F.ElectricLogic) is not None,
        )

    @classmethod
    def add_to_all_suitable_modules(cls, node: Node, params: Params | None = None):
        layout = cls(params)
        candidates = cls.find_module_candidates(node)
        for c in candidates:
            logger.debug(f"Adding {cls.__name__} to {c}")
            c.add(F.has_pcb_layout_defined(layout))
        return candidates



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import re
import subprocess
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent, indent

import typer

from faebryk.libs.tools.typer import typer_callback


@dataclass
class CTX: ...


def get_ctx(ctx: typer.Context) -> "CTX":
    return ctx.obj


@typer_callback(None)
def main(ctx: typer.Context):
    """
    Can be called like this: > faebryk refactor
    """
    pass


pyname = r"[_a-zA-Z][_a-zA-Z0-9]*"


@main.command()
def libtof(ctx: typer.Context, root: Path):
    file_paths = list(root.rglob("**/*.py"))
    print(f"Found {len(file_paths)} files in path.")

    detection_pattern = re.compile(r"^from faebryk.library.[^_]")

    refactor_files = [
        path
        for path in file_paths
        if not path.stem.startswith("_")
        and any(
            detection_pattern.match(line)
            for line in path.read_text(encoding="utf-8").splitlines()
        )
    ]

    print(f"Found {len(refactor_files)} files to refactor.")

    # TO match:
    # from faebryk.library.has_kicad_footprint import has_kicad_footprint
    # from faebryk.library.has_simple_value_representation import (
    #     has_simple_value_representation,
    # )

    import_pattern = re.compile(
        r"^from faebryk.library.([^_][^ ]*) import ("
        f"{pyname}$"
        "|"
        f"\\([\n ]*{pyname},[\n ]*\\)$"  # multiline import
        r")",
        re.MULTILINE,
    )

    def refactor_file(path: Path):
        text = path.read_text(encoding="utf-8")
        import_symbols = [m[0] for m in import_pattern.findall(text)]
        text = import_pattern.subn("import faebryk.library._F as F", text, count=1)[0]
        text = import_pattern.sub("", text)

        for sym in import_symbols:
            if re.search(rf"from faebryk.library.{sym} import {sym} as", text):
                print(f"Warning: skipping {sym} in {path}")
                continue

            text = re.sub(rf"^\s*from faebryk.library.{sym} import {sym}$", "", text)
            text = re.sub(
                rf"^\s*from faebryk.library.{sym} import \(\n\s*{sym},\n\)$", "", text
            )
            text = re.sub(rf"(?<!F\.)\b{sym}\b", f"F.{sym}", text)

        # print(path, import_symbols)
        path.write_text(text, encoding="utf-8")

    for path in refactor_files:
        refactor_file(path)

    # Run ruff
    subprocess.check_call(["ruff", "check", "--fix", root])


maybe_f = r"(?:F\.)?"


@main.command()
def fabll(ctx: typer.Context, root: Path):
    file_paths = list(root.rglob("**/*.py"))
    print(f"Found {len(file_paths)} files in path.")

    types = r"(?:IF|NODE|PARAM)"
    ano_class = rf"class _{types}s\("
    # detection_pattern = re.compile(ano_class)

    refactor_files = file_paths
    # refactor_files = [
    #    path
    #    for path in file_paths
    #    if not path.stem.startswith("_") and detection_pattern.search(
    # path.read_text(encoding="utf-8"))
    # ]

    print(f"Found {len(refactor_files)} files to refactor.")

    holder_header = rf"[ ]*{ano_class}.*?\):\n"
    holder_pattern = re.compile(
        rf"({holder_header})(.*?)\n\n", re.MULTILINE | re.DOTALL
    )

    instance_holder = re.compile(rf"self.{types}s = _{types}s\(self\)", re.MULTILINE)
    holder_attr = re.compile(rf"\.{types}s\.")

    def refactor_file(path: Path):
        text = path.read_text(encoding="utf-8")

        print(path, "=" * 40)

        def process(m: re.Match) -> str:
            # print("Block", "|||")
            header, fields = m.groups()
            fields: str = fields + "\n"
            out = []
            for f in fields.split(")\n"):
                if not f:
                    continue
                f += ")"
                # print(f)
                # print(indent("|\nv", " " * 12))

                if "times" in f:
                    f = re.sub(r"\n\s*", "", f)
                    f = f.replace(",)", ")")

                    # case 3: screw_holes = times(3, lambda: F.Mounting_Hole())
                    f = re.sub(r"times\((\d+),", r"L.list_field(\1,", f)

                    # case 5:  leds = times(\n self.pixels.value, \n ...,\n)
                    f = re.sub(
                        rf"\s*({pyname}) = (times\(.*\))",
                        r"@L.rt_field\ndef \1(self):\n    return \2",
                        f,
                    )

                # case 1: uart_in = F.UART_Base()
                # case 4:  if buffered:\n power_data = F.ElectricPower()
                f = re.sub(rf" = ({maybe_f}{pyname})\(\)", r": \1", f)

                # case 2: fan_power_switch = F.PowerSwitchMOSFET(lowside=True, ...)
                f = re.sub(
                    rf" = ({maybe_f}{pyname})\((.*)\)", r" = L.f_field(\1)(\2)", f
                )

                # print(f)
                # print("--")
                out.append(dedent(f))

            outstr = indent("\n".join(out), " " * 4)
            # print(outstr)
            return outstr

        # Holders -------------------------------------
        text = holder_pattern.sub(process, text)
        text = instance_holder.sub("", text)
        text = holder_attr.sub(".", text)

        # Init ----------------------------------------
        # convert empty constructor to __preinit__
        text = re.sub(
            r"def __init__\(self\).*?:(.*?)super\(\).__init__\(\)",
            r"def __preinit__(self):\1pass",
            text,
            flags=re.DOTALL | re.MULTILINE,
        )

        # remove -> None from init
        text = re.sub(
            r"(def __init__\(self.*?\)).*?:",
            r"\1:",
            text,
        )

        def process_constructor(m: re.Match) -> str:
            str_: str = m.group(0)
            args = m.group(1)
            out = str_
            prefix = re.match(r"^\s*", str_).group(0)

            # find names of args
            arg_names = re.findall(rf",\s*({pyname})\s*(?:[:,]|$)", args)
            for arg in arg_names:
                out += indent(f"\n    self._{arg} = {arg}", prefix)

            out += indent("\n\ndef __preinit__(self):\n    pass", prefix)
            return out

        # split args of constructor into init and pre_init
        text = re.sub(
            r"^[ ]*def __init__\(self(, .*?)\):.*?super\(\).__init__\(\)",
            process_constructor,
            text,
            flags=re.DOTALL | re.MULTILINE,
        )

        # Done ----------------------------------------
        text = "\n".join(line.rstrip() for line in text.splitlines())
        text = text.strip() + "\n"
        path.write_text(text, encoding="utf-8")

    for path in refactor_files:
        refactor_file(path)


if __name__ == "__main__":
    main()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class EEPROM(Module):
    """
    Generic EEPROM module with F.I2C interface.
    """

    def set_address(self, addr: int):
        """
        Configure the address of the EEPROM by setting the address pins.
        """
        assert addr < (1 << len(self.address))

        for i, e in enumerate(self.address):
            e.set(addr & (1 << i) != 0)

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    memory_size = L.p_field(
        units=P.bit,
        likely_constrained=True,
        domain=L.Domains.Numbers.NATURAL(),
        soft_set=L.Range(128 * P.bit, 1024 * P.kbit),
    )

    power: F.ElectricPower
    i2c: F.I2C
    write_protect: F.ElectricLogic
    address = L.list_field(3, F.ElectricLogic)

    # ----------------------------------------
    #                traits
    # ----------------------------------------

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



import pytest

from faebryk.libs.units import (
    P,
    Unit,
    UnitCompatibilityError,
    assert_compatible_units,
)


class DummyHasUnit:
    def __init__(self, units: Unit):
        self.units = units


def test_assert_compatible_units_empty_list():
    """Test that empty list raises ValueError"""
    with pytest.raises(ValueError):
        assert_compatible_units([])


def test_assert_compatible_units_single_item():
    """Test that single item list returns its unit"""
    item = DummyHasUnit(P.meter)
    result = assert_compatible_units([item])
    assert result == P.meter


def test_assert_compatible_units_compatible():
    """Test that compatible units pass validation and return first unit"""
    items = [
        DummyHasUnit(P.meter),
        DummyHasUnit(P.kilometer),
        DummyHasUnit(P.centimeter),
    ]
    result = assert_compatible_units(items)
    assert result == P.meter


def test_assert_compatible_units_incompatible():
    """Test that incompatible units raise UnitCompatibilityError"""
    items = [
        DummyHasUnit(P.meter),
        DummyHasUnit(P.second),
    ]
    with pytest.raises(UnitCompatibilityError) as exc_info:
        assert_compatible_units(items)

    assert len(exc_info.value.incompatible_items) == 2


def test_assert_compatible_units_with_derived():
    """Test that derived units are handled correctly and return first unit"""
    items = [
        DummyHasUnit(P.meter / P.second),
        DummyHasUnit(P.kilometer / P.hour),
    ]
    result = assert_compatible_units(items)
    assert result == P.meter / P.second


def test_assert_compatible_units_with_incompatible_derived():
    """Test that incompatible derived units raise UnitCompatibilityError"""
    items = [
        DummyHasUnit(P.meter / P.second),
        DummyHasUnit(P.meter * P.second),
    ]
    with pytest.raises(UnitCompatibilityError):
        assert_compatible_units(items)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod

import faebryk.library._F as F
from faebryk.core.module import Module


class can_attach_to_footprint(Module.TraitT):
    @abstractmethod
    def attach(self, footprint: F.Footprint): ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class USBLC6_2P6(Module):
    """
    Low capacitance TVS diode array (for USB2.0)
    """

    # interfaces
    usb: F.USB2_0

    @L.rt_field
    def attach_to_footprint(self):
        x = self
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": x.usb.usb_if.d.p.line,
                "2": x.usb.usb_if.buspower.lv,
                "3": x.usb.usb_if.d.n.line,
                "4": x.usb.usb_if.d.n.line,
                "5": x.usb.usb_if.buspower.hv,
                "6": x.usb.usb_if.d.p.line,
            }
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://datasheet.lcsc.com/lcsc/2108132230_TECH-PUBLIC-USBLC6-2P6_C2827693.pdf"
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.exporters.pcb.routing.util import (
    DEFAULT_TRACE_WIDTH,
    DEFAULT_VIA_SIZE_DRILL,
    Path,
    Route,
    get_internal_nets_of_node,
    get_pads_pos_of_mifs,
    get_routes_of_pad,
    group_pads_that_are_connected_already,
)
from faebryk.libs.geometry.basic import Geometry

logger = logging.getLogger(__name__)


class has_pcb_routing_strategy_via_to_layer(F.has_pcb_routing_strategy.impl()):
    def __init__(self, layer: str, vec: Geometry.Point2D):
        super().__init__()
        self.vec = vec
        self.layer = layer

    def calculate(self, transformer: PCB_Transformer):
        layer = transformer.get_layer_id(self.layer)

        node = self.obj
        nets = get_internal_nets_of_node(node)

        logger.debug(f"Routing {node} {'-' * 40}")

        def get_route_for_net(net: F.Net, mifs) -> Route | None:
            net_name = net.get_trait(F.has_overriden_name).get_name()

            pads = get_pads_pos_of_mifs(mifs)
            pad_groups = group_pads_that_are_connected_already(pads)

            # Get pad representatives in groups
            # Filter groups that already are connected to VIAs
            pads_filtered = {
                pad: pads[pad]
                for group in pad_groups
                if (pad := next(iter(group)))
                and not any(
                    isinstance(path_elem, Path.Via)
                    for r in get_routes_of_pad(pad)
                    for path_elem in r.path.path
                )
            }

            logger.debug(f"Routing net {net_name} with pads: {pads_filtered}")

            path = Path()

            for _, pos in pads_filtered.items():
                # No need to add via if on same layer already
                pad_layer = pos[3]
                if pad_layer == layer:
                    continue
                via_pos: Geometry.Point = Geometry.add_points(pos, self.vec)
                path.add(Path.Via(via_pos, size_drill=DEFAULT_VIA_SIZE_DRILL))
                path.add(Path.Line(DEFAULT_TRACE_WIDTH, pad_layer, pos, via_pos))

            path.add(
                Path.Zone(
                    layer,
                    transformer.get_net_obj_bbox(
                        transformer.get_net(net),
                        layer=self.layer,
                        tolerance=Geometry.distance_euclid((0, 0, 0, 0), self.vec),
                    ),
                )
            )

            return Route(pads, path)

        return [
            route
            for net, mifs in nets.items()
            if net and (route := get_route_for_net(net, mifs))
        ]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.libs.library import L  # noqa: F401

logger = logging.getLogger(__name__)


class Diodes_Incorporated_AP2552W6_7(F.Diodes_Incorporated_AP255x_x):
    """
    Power Distribution Switch. Mostly used in USB applications.
    - Active low enable
    - Current limit during overcurrent

    2.7V~5.5V 70mΩ 2.1A SOT-26
    """

    explicit_part = L.f_field(F.has_explicit_part.by_supplier)("C441824")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.util import not_none


class Pad(ModuleInterface):
    net: F.Electrical
    pcb: ModuleInterface

    def attach(self, intf: F.Electrical):
        self.net.connect(intf)
        intf.add(F.has_linked_pad_defined(self))

    @staticmethod
    def find_pad_for_intf_with_parent_that_has_footprint_unique(
        intf: ModuleInterface,
    ) -> "Pad":
        pads = Pad.find_pad_for_intf_with_parent_that_has_footprint(intf)
        if len(pads) != 1:
            raise ValueError
        return next(iter(pads))

    @staticmethod
    def find_pad_for_intf_with_parent_that_has_footprint(
        intf: ModuleInterface,
    ) -> list["Pad"]:
        # This only finds directly attached pads
        # -> misses from parents / children nodes
        # if intf.has_trait(F.has_linked_pad):
        #     return list(intf.get_trait(F.has_linked_pad).get_pads())

        # This is a bit slower, but finds them all
        _, footprint = F.Footprint.get_footprint_of_parent(intf)
        pads = [
            pad
            for pad in footprint.get_children(direct_only=True, types=Pad)
            if pad.net.is_connected_to(intf)
        ]
        return pads

    def get_fp(self) -> F.Footprint:
        return not_none(self.get_parent_of_type(F.Footprint))



import itertools
import logging
import re
import subprocess
import sys
import textwrap
from enum import StrEnum, auto
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Callable, Iterator, cast

import caseconverter
import questionary
import rich
import typer
from cookiecutter.exceptions import OutputDirExistsException
from cookiecutter.main import cookiecutter
from more_itertools import first
from rich.table import Table

from atopile import errors, version
from atopile.address import AddrStr
from atopile.config import PROJECT_CONFIG_FILENAME, config
from atopile.telemetry import log_to_posthog
from faebryk.libs.github import (
    GithubCLI,
    GithubCLINotFound,
    GithubRepoAlreadyExists,
    GithubRepoNotFound,
    GithubUserNotLoggedIn,
)
from faebryk.libs.logging import rich_print_robust
from faebryk.libs.util import get_code_bin_of_terminal, try_or

if TYPE_CHECKING:
    import git

    from faebryk.libs.picker.api.api import Component

# Set up logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


create_app = typer.Typer(
    rich_markup_mode="rich", help="Create projects / build targets / components"
)


def help(text: str) -> None:  # pylint: disable=redefined-builtin
    """Print help text."""
    rich_print_robust("\n" + textwrap.dedent(text).strip() + "\n")


def _stuck_user_helper() -> Iterator[bool]:
    """Figure out if a user is stuck and help them exit."""
    threshold = 5
    for i in itertools.count():
        if i >= threshold:
            if questionary.confirm("Are you trying to exit?").unsafe_ask():
                rich_print_robust("No worries! Try Ctrl+C next time!")
                exit(0)
            threshold += 5
        yield True


stuck_user_helper_generator = _stuck_user_helper()


def _in_git_repo(path: Path) -> bool:
    """Check if the current directory is in a git repo."""
    import git

    try:
        git.Repo(path)
    except git.InvalidGitRepositoryError:
        return False
    return True


def query_helper[T: str | Path | bool](
    prompt: str,
    type_: type[T],
    clarifier: Callable[[Any], T] = lambda x: x,
    upgrader: Callable[[T], T] = lambda x: x,
    upgrader_msg: str | None = None,
    validator: str | Callable[[T], bool] | None = None,
    validation_failure_msg: str | None = "Value [cyan]{value}[/] is invalid",
    default: T | None = None,
    pre_entered: T | None = None,
    validate_default: bool = True,
) -> T:
    """Query a user for input."""
    rich_print_robust(prompt)

    # Check the default value
    if default is not None:
        if not isinstance(default, type_):
            raise ValueError(f"Default value {default} is not of type {type_}")

    # Make a querier
    if type_ is str:

        def querier() -> str:  # type: ignore
            return questionary.text(
                "",
                default=str(default or ""),
            ).unsafe_ask()

    elif type_ is Path:

        def querier() -> Path:  # type: ignore
            return Path(
                questionary.path(
                    "",
                    default=str(default or ""),
                ).unsafe_ask()
            )

    elif type_ is bool:
        assert default is None or isinstance(default, bool)
        if default is None:
            default = True  # type: ignore

        def querier() -> bool:
            return questionary.confirm(
                "",
                default=default,  # type: ignore
            ).unsafe_ask()

    else:
        raise ValueError(f"Unsupported query type: `{type_}`")

    # Ensure a validator
    # Default the validator to a regex match if it's a str
    if validator is None:

        def validator_func(value: T) -> bool:
            return True

    elif isinstance(validator, str):

        def validator_func(value: T) -> bool:
            return bool(re.match(validator, value))  # type: ignore

        if not validation_failure_msg:
            validation_failure_msg = f'Value must match regex: `r"{validator}"`'

    else:
        validator_func = validator  # type: ignore

    # Ensure the default provided is valid
    if default is not None and validate_default:
        if not validator_func(clarifier(default)):
            logger.debug(f"Default value {default} is invalid")

        if clarifier(default) != upgrader(clarifier(default)):
            logger.debug(f"Default value {default} doesn't meet best-practice")

        default = upgrader(clarifier(default))

    # When running non-interactively, we expect the value to be provided
    # at the command level, so we don't need to query the user for it
    # Validate and return the pre-entered value
    if not config.interactive:
        if pre_entered is None:
            raise ValueError("Value is required. Check at command level.")

        if not validator_func(pre_entered):
            if validation_failure_msg:
                msg = validation_failure_msg.format(value=pre_entered)
            else:
                msg = f"Value {pre_entered} is invalid"
            raise errors.UserException(msg)

        return pre_entered

    # Pre-entered values are expected to skip the query for a value in the first place
    # but progress through the validator and upgrader if we're running interactively
    value: T | None = pre_entered

    for _ in stuck_user_helper_generator:
        if value is None:
            value = clarifier(querier())  # type: ignore
        assert isinstance(value, type_)

        if (proposed_value := upgrader(value)) != value:
            if upgrader_msg:
                rich_print_robust(upgrader_msg.format(proposed_value=proposed_value))

            rich_print_robust(f"Use [cyan]{proposed_value}[/] instead?")
            if questionary.confirm("").unsafe_ask():
                value = proposed_value

        if not validator_func(value):
            if validation_failure_msg:
                rich_print_robust(validation_failure_msg.format(value=value))
            value = None
            continue

        return value

    raise RuntimeError("Unclear how we got here")


PROJECT_NAME_REQUIREMENTS = (
    "Project name must start with a letter and contain only letters, numbers, dashes"
    " and underscores. It will be used for the project directory and name on GitHub"
)


def setup_github(
    project_path: Path,
    gh_cli: GithubCLI,
    repo: "git.Repo",
):
    import git

    github_username = gh_cli.get_usernames()

    use_existing_repo = query_helper(
        "Use an existing GitHub repo?",
        bool,
        default=False,
    )

    if use_existing_repo:

        def _check_repo_and_add_remote(repo_id: str) -> bool:
            try:
                url = gh_cli.get_repo_url(repo_id)
                repo.create_remote("origin", url)
                rich_print_robust(f"Added remote origin: {url}")
                # Try to push, but don't fail validation if it doesn't work
                # (e.g. repo not empty, or other reasons)
                try:
                    rich_print_robust(
                        f"Attempting to push initial commit to"
                        f" {repo.active_branch.name}..."
                    )
                    repo.git.push("-u", "origin", repo.active_branch.name)
                    rich_print_robust("[green]Pushed successfully![/]")
                except git.GitCommandError as e:
                    rich_print_robust(
                        f"[yellow]Could not push to remote:[/yellow] {e.stderr.strip()}"
                    )
                    rich_print_robust("You may need to push manually.")
                return True
            except GithubRepoNotFound:
                rich_print_robust(f"[red]Repository {repo_id} not found on GitHub.[/]")
                return False
            except git.GitCommandError as e:
                # This might happen if remote 'origin' already exists
                rich_print_robust(
                    f"[red]Failed to add remote:[/red] {e.stderr.strip()}"
                )
                return False
            except KeyboardInterrupt:
                rich_print_robust("[red]Aborted.[/red]")
                return False
            except Exception as e:
                rich_print_robust(f"[red]An unexpected error occurred:[/red] {e}")
                return False

        query_helper(
            ":rocket: What's the [cyan]repo's org and project name?[/]\n",
            str,
            default=f"{github_username[0]}/{project_path.name}",
            validator=_check_repo_and_add_remote,
            validation_failure_msg="Remote could not be added: {value}",
            validate_default=False,
        )
    else:

        def _create_repo_and_add_remote(repo_id: str) -> bool:
            try:
                # This will create the repo, add remote 'origin'
                #   and push the current dir
                # We are currently in project_path, which is the root of
                #   the new git repo
                # Ask for visibility
                visibility = questionary.select(
                    "Choose repository visibility:",
                    choices=["public", "private"],
                    default="public",
                ).unsafe_ask()

                repo_url = gh_cli.create_repo(
                    repo_id,
                    visibility=visibility,
                    add_remote=True,
                    path=project_path,
                )
                rich_print_robust(
                    f"[green]Successfully created repository {repo_url} and"
                    " pushed initial commit![/]"
                )
                return True
            except GithubRepoAlreadyExists:
                rich_print_robust(
                    f"[red]Repository {repo_id} already exists on GitHub.[/]"
                )
                # We could offer to use it, but for now, let's just fail validation
                return False
            except git.GitCommandError as e:
                # This might happen if the push fails for some reason
                rich_print_robust(
                    f"[red]Failed during git operation (e.g. push):[/red]"
                    f" {e.stderr.strip()}"
                )
                return False
            except KeyboardInterrupt:
                rich_print_robust("[red]Aborted.[/red]")
                return False
            except Exception as e:
                rich_print_robust(f"[red]An unexpected error occurred:[/red] {e}")
                return False

        query_helper(
            ":rocket: Choose a [cyan]repo org and name:[/]\n",
            str,
            default=f"{github_username[0]}/{project_path.name}",
            validator=_create_repo_and_add_remote,
            validation_failure_msg="Remote could not be added: {value}",
            validate_default=False,
        )


@create_app.command()
@log_to_posthog("cli:create_project_end")
def project(path: Annotated[Path | None, typer.Option()] = None):
    """
    Create a new ato project.
    """
    TEMPLATE_DIR = Path(__file__).parent.parent / "project-template"

    extra_context = {
        "__ato_version": version.get_installed_atopile_version(),
        "__python_path": sys.executable,
    }

    if path is None:
        if config.interactive:
            path = query_helper(
                ":rocket: Where should we create the project?",
                type_=Path,
                default=Path.cwd(),
                pre_entered=path,
                validator=lambda x: x.is_dir(),
            )
        else:
            path = Path.cwd()

    logging.info("Running cookie-cutter on the template")
    try:
        project_path = Path(
            cookiecutter(
                str(TEMPLATE_DIR),
                output_dir=str(path),
                no_input=not config.interactive,
                extra_context=dict(
                    filter(lambda x: x[1] is not None, extra_context.items())
                ),
            )
        )
    except OutputDirExistsException as e:
        raise errors.UserException(
            "Directory already exists. Please choose a different name."
        ) from e

    try:
        import git

        no_git = False
    except ImportError as e:
        # catch no git executable
        if "executable" not in e.msg:
            raise
        no_git = True

    if not no_git:
        # check if already in a git repo
        create_git_repo = not _in_git_repo(project_path)

        # check if gh binary is available
        gh_cli = try_or(GithubCLI, catch=(GithubCLINotFound, GithubUserNotLoggedIn))

        # git repo
        if create_git_repo:
            logging.info("Initializing git repo")
            repo = git.Repo.init(project_path)
            repo.git.add(A=True, f=True)
            try:
                repo.git.commit(m="Initial commit")
            except git.GitCommandError as e:
                if "Author identity unknown" in e.stderr:
                    rich_print_robust(
                        "[yellow]Warning: Author identity unknown. "
                        "Staged but not committed.[/yellow]"
                    )
                else:
                    raise

        create_github_repo = False
        if config.interactive and gh_cli and create_git_repo:
            create_github_repo = query_helper(
                "Host this project on GitHub? :octopus::cat:",
                bool,
                default=False,
            )

        # Github repo
        if create_github_repo:
            assert gh_cli is not None
            try:
                setup_github(project_path, gh_cli, repo)
            except Exception:
                rich_print_robust("[red]Creating GitHub repo interrupted.[/red]")
                return

    # Wew! New repo created!
    rich_print_robust(
        f':sparkles: [green]Created new project "{project_path.name}"![/] :sparkles:'
    )

    # check if running in vscode / cursor terminal
    if code_bin := get_code_bin_of_terminal():
        # open in vscode / cursor
        subprocess.Popen([code_bin, project_path])
    else:
        rich_print_robust(f" \n[cyan]cd {project_path.relative_to(Path.cwd())}[/cyan]")


@create_app.command("build-target")
@log_to_posthog("cli:create_build_target_end")
def build_target(
    build_target: Annotated[str | None, typer.Option()] = None,
    file: Annotated[Path | None, typer.Option()] = None,
    module: Annotated[str | None, typer.Option()] = None,
    backup_name: Annotated[str | None, typer.Argument()] = None,
):
    """
    Create a new build configuration.
    - adds entry to ato.yaml
    - creates a new directory in layout
    """
    config.apply_options(None)

    try:
        src_path = config.project.paths.src
        config.project_dir  # touch property to ensure config's loaded from a project
    except ValueError:
        raise errors.UserNoProjectException()

    if backup_name is None:
        if build_target:
            backup_name = build_target
        elif file:
            backup_name = file.name.split(".", 1)[0]
        elif module:
            backup_name = module

    # If we're running non-interactively, all details must be provided
    if not config.interactive and not all([build_target, file, module]):
        raise errors.UserException(
            "--build-target, --file, and --module must all be"
            " provided when running non-interactively."
        )

    def _check_build_target_name(value: str) -> bool:
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9_-]*$", value):
            rich_print_robust(
                "[red]Build-target names must start with a letter and"
                " contain only letters, numbers, dashes and underscores.[/]"
            )
            return False

        if value in config.project.builds:
            rich_print_robust(f"[red]Build-target `{value}` already exists[/]")
            return False

        return True

    build_target = query_helper(
        ":rocket: What's the [cyan]build-target[/] name?",
        type_=str,
        upgrader=caseconverter.kebabcase,
        upgrader_msg="We recommend using kebab-case for build-target names.",
        validator=_check_build_target_name,
        validation_failure_msg="",
        pre_entered=build_target,
        default=caseconverter.kebabcase(backup_name) if backup_name else None,
    )

    def _file_clarifier(value: Path) -> Path:
        if not value.is_absolute():
            # If it's a relative path wrt ./ (cwd) then respect that
            # else assume it's relative to the src directory
            if str(value).startswith("./") or str(value).startswith(".\\"):
                value = Path(value).resolve()
            else:
                value = src_path / value

        return value

    def _file_updator(value: Path) -> Path:
        value = value.with_stem(caseconverter.snakecase(value.stem))
        if value.suffix != ".ato":
            # Allow dots in filenames
            value = value.with_suffix(value.suffix + ".ato")
        return value

    def _file_validator(f: Path) -> bool:
        if f.is_dir():
            rich_print_robust(f"{f} is a directory")
            return False

        if f.suffix != ".ato":
            rich_print_robust(f"{f} must end in .ato")
            return False

        try:
            f.relative_to(src_path)
        except ValueError:
            rich_print_robust(f"{f} is outside the project's src dir")
            return False

        return True

    file = query_helper(
        ":rocket: What [cyan]file[/] should we add the module to?",
        type_=Path,
        clarifier=_file_clarifier,
        upgrader=_file_updator,
        upgrader_msg=(
            "We recommend using snake_case for file names, and it must end in .ato"
        ),
        validator=_file_validator,
        validation_failure_msg="",
        pre_entered=file,
        default=Path(caseconverter.snakecase(backup_name or build_target) + ".ato"),
    )

    module = query_helper(
        ":rocket: What [cyan]module[/] should we add to the file?",
        type_=str,
        validator=str.isidentifier,
        validation_failure_msg="",
        upgrader=caseconverter.pascalcase,
        upgrader_msg="We recommend using pascal-case for module names.",
        pre_entered=module,
        default=caseconverter.pascalcase(backup_name or build_target),
    )

    logger.debug(f"Creating build-target with {build_target=}, {file=}, {module=}")

    # Update project config
    logger.info(
        f"Adding build-target to {config.project.paths.root / PROJECT_CONFIG_FILENAME}"
    )

    def add_build_target(config_data: dict, new_data: dict):
        config_data["builds"][build_target] = new_data
        return config_data

    config.update_project_settings(
        add_build_target, {"entry": str(AddrStr.from_parts(file, module))}
    )

    # Create or add to file
    module_text = f"module {module}:\n    pass\n"

    if file.is_file():  # exists and is a file
        with file.open("a", encoding="utf-8") as f:
            f.write("\n")
            f.write(module_text)

    else:
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(module_text, encoding="utf-8")

    rich_print_robust(
        ":sparkles: Successfully created a new build configuration "
        f"[cyan]{build_target}[/] at [cyan]{file}[/]! :sparkles:"
    )


class ComponentType(StrEnum):
    ato = auto()
    fab = auto()


@create_app.command()
@log_to_posthog("cli:create_component_end")
def part(
    search_term: Annotated[str | None, typer.Option("--search", "-s")] = None,
    accept_single: Annotated[bool, typer.Option("--accept-single", "-a")] = False,
):
    """Create a new component."""
    from faebryk.libs.picker.api.api import ApiHTTPError
    from faebryk.libs.picker.api.picker_lib import _extract_numeric_id, client
    from faebryk.libs.picker.lcsc import download_easyeda_info

    config.apply_options(None)

    # Find a component --------------------------------------------------------

    component: Component | None = None

    for _ in stuck_user_helper_generator:
        if not search_term:
            search_term = questionary.text(
                "Search for a component (Part Number or LCSC ID):"
            ).unsafe_ask()
            assert search_term is not None

        try:
            lcsc_id = _extract_numeric_id(search_term)
        except ValueError:
            lcsc_id = None

        try:
            if lcsc_id:
                components = client.fetch_part_by_lcsc(lcsc_id)
            else:
                if ":" in search_term:
                    mfr, search_term = search_term.split(":", 1)
                else:
                    # TODO: remove this once we have a fuzzy search
                    mfr = questionary.text("Enter the manufacturer").unsafe_ask()
                components = client.fetch_part_by_mfr(mfr, search_term)
        except ApiHTTPError as e:
            if e.response.status_code == 404:
                components = []
            else:
                raise

        if len(components) == 0:
            rich_print_robust(f'No components found for "{search_term}"')
            search_term = None
            continue

        component_table = Table()
        component_table.add_column("Manufacturer")
        component_table.add_column("Part Number")
        component_table.add_column("Description")
        component_table.add_column("Supplier ID")
        component_table.add_column("Stock")

        components = sorted(components, key=lambda c: c.stock, reverse=True)

        for component in components:
            component_table.add_row(
                component.manufacturer_name,
                component.part_number,
                component.description,
                component.lcsc_display,
                str(component.stock),
            )

        rich.print(component_table)

        if len(components) == 1 and accept_single:
            component = first(components)
        else:
            choices = [
                {
                    "name": f"{component.manufacturer_name} {component.part_number}"
                    f" ({component.lcsc_display})",
                    "value": component,
                }
                for component in components
            ] + [{"name": "Search again...", "value": None}]

            component = questionary.select(
                "Select a component", choices=choices
            ).unsafe_ask()

        if component is not None:
            break

        # Reset the input terms to start over if we didn't find what we're looking for
        search_term = None

    # We have a component -----------------------------------------------------
    assert component is not None
    from faebryk.libs.part_lifecycle import PartIsNotAutoGenerated, PartLifecycle

    try:
        epart = download_easyeda_info(component.lcsc_display, get_model=True)
        apart = PartLifecycle.singleton().library.ingest_part_from_easyeda(epart)
    except PartIsNotAutoGenerated as e:
        raise errors.UserException(
            f"Part `{e.part.path}` already exists and is manually modified."
        ) from e
    except Exception as e:
        raise errors.UserException(str(e)) from e

    rich_print_robust(f":sparkles: Created {apart.identifier} at {apart.path} !")


@create_app.command(deprecated=True)
def component(
    search_term: Annotated[str | None, typer.Option("--search", "-s")] = None,
):
    return part(search_term)


@create_app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    if ctx.resilient_parsing:
        return

    if not ctx.invoked_subcommand:
        commands = cast(dict, ctx.command.commands)  # type: ignore  # commands is an attribute of the context
        command_name = questionary.select(
            "What would you like to create?",
            choices=[n for n, c in commands.items() if not c.hidden],
        ).unsafe_ask()

        assert command_name in commands

        # Run the command
        ctx.invoke(commands[command_name].callback)


if __name__ == "__main__":
    create_app()  # pylint: disable=no-value-for-parameter



from faebryk.libs.util import has_instance_settable_attr, write_only_property


class SomeClass:
    def __init__(self):
        self.regular_attr = "initial"
        self._rw_value = "initial"

    @property
    def read_only_prop(self):
        return "read_only"

    @write_only_property
    def write_only_prop(self, value):
        self._wo_value = value

    @property
    def read_write_prop(self):
        return self._rw_value

    @read_write_prop.setter
    def read_write_prop(self, value):
        self._rw_value = value


def test_set_regular_attribute():
    """Test setting a regular attribute"""
    obj = SomeClass()
    assert has_instance_settable_attr(obj, "regular_attr")


def test_set_new_attribute():
    """Test setting a new attribute (should fail)"""
    obj = SomeClass()
    assert not has_instance_settable_attr(obj, "new_attr")


def test_set_read_only_property():
    """Test setting a read-only property (should fail)"""
    obj = SomeClass()
    assert not has_instance_settable_attr(obj, "read_only_prop")


def test_set_read_write_property():
    """Test setting a read-write property"""
    obj = SomeClass()
    assert has_instance_settable_attr(obj, "read_write_prop")


def test_set_nonexistent_attribute():
    """Test setting a nonexistent attribute"""
    obj = SomeClass()
    assert not has_instance_settable_attr(obj, "nonexistent_attr")


def test_set_class_level_attribute():
    """Test setting a class-level attribute"""

    class ClassWithClassAttr:
        class_attr = "initial"

    obj = ClassWithClassAttr()
    assert not has_instance_settable_attr(obj, "class_attr")


def test_set_inherited_property():
    """Test setting an inherited property"""

    class ChildClass(SomeClass):
        pass

    obj = ChildClass()
    assert has_instance_settable_attr(obj, "read_write_prop")


def test_set_none_object():
    """Test setting attribute on None (should fail)"""
    assert not has_instance_settable_attr(None, "attr")


def test_set_primitive_types():
    """Test setting attributes on primitive types (should fail)"""
    assert not has_instance_settable_attr(42, "attr")
    assert not has_instance_settable_attr("string", "attr")


def test_set_with_property_no_setter():
    """Test setting a property that has no setter"""

    class NoSetterClass:
        @property
        def prop(self):
            return "value"

    obj = NoSetterClass()
    assert not has_instance_settable_attr(obj, "prop")


def test_write_only_property():
    """Test setting a write-only property"""
    obj = SomeClass()
    assert has_instance_settable_attr(obj, "write_only_prop")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_footprint_defined(F.has_footprint_impl):
    def __init__(self, fp: F.Footprint) -> None:
        super().__init__()
        self.fp = fp

    def on_obj_set(self):
        self.set_footprint(self.fp)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class Wuxi_I_core_Elec_AiP74LVC1T45GB236_TR(Module):
    """
    Single channel bidirectional buffer.
    1.2V-5.5V logic levels.
    SOT-23-6
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    direction: F.ElectricLogic
    port_a: F.ElectricLogic
    power_a: F.ElectricPower
    port_b: F.ElectricLogic
    power_b: F.ElectricPower

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    explicit_part = L.f_field(F.has_explicit_part.by_mfr)(
        "Wuxi I-core Elec", "AiP74LVC1T45GB236.TR"
    )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.port_a, self.port_b)

    @L.rt_field
    def decoupled(self):
        return F.can_be_decoupled_rails(self.power_a, self.power_b)

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.port_a.line: ["A"],
                self.port_b.line: ["B"],
                self.direction.line: ["DIR"],
                self.power_a.lv: ["GND"],
                self.power_a.hv: ["VCC(A)"],
                self.power_b.hv: ["VCC(B)"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)
        F.ElectricLogic.connect_all_module_references(
            self, exclude=[self.power_a, self.port_a, self.direction]
        )
        F.ElectricLogic.connect_all_module_references(
            self, exclude=[self.power_b, self.port_b]
        )

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.power_a.voltage.constrain_subset(L.Range(1.2 * P.V, 5.5 * P.V))
        self.power_b.voltage.constrain_subset(L.Range(1.2 * P.V, 5.5 * P.V))

        # FIXME
        # self.power_a.decoupled.decouple().capacitance.constrain_subset(
        #     L.Range.from_center(100 * P.nF, 10 * P.nF)
        # )
        # self.power_b.decoupled.decouple().capacitance.constrain_subset(
        #     L.Range.from_center(100 * P.nF, 10 * P.nF)
        # )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class can_attach_via_pinmap_equal(F.can_attach_via_pinmap.impl()):
    def attach(self, pinmap: dict[str, F.Electrical]):
        pin_list = {
            v: k for k, v in self.obj.get_trait(F.has_equal_pins).get_pin_map().items()
        }
        for no, intf in pinmap.items():
            pin_list[no].attach(intf)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import math

logger = logging.getLogger(__name__)


def sine_on_interval(
    interval: tuple[float, float],
) -> tuple[float, float]:
    """
    Computes the overall sine range on the given x-interval.

    The extreme values occur either at the endpoints or at turning points
    of sine (x = π/2 + π*k).
    """
    start, end = interval
    if start > end:
        raise ValueError("Invalid interval: start must be <= end")
    if math.isinf(start) or math.isinf(end):
        return (-1, 1)
    if end - start > 2 * math.pi:
        return (-1, 1)

    # Evaluate sine at the endpoints
    xs = [start, end]

    # Include turning points within the interval
    k_start = math.ceil((start - math.pi / 2) / math.pi)
    k_end = math.floor((end - math.pi / 2) / math.pi)
    for k in range(k_start, k_end + 1):
        xs.append(math.pi / 2 + math.pi * k)

    sine_values = [math.sin(x) for x in xs]
    return (min(sine_values), max(sine_values))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class SP3243E(Module):
    """
    250Kbps RS232 Transceiver 3tx/5rx SSOP28
    """

    def enable_chip(self):
        self.enable.set(on=True)

    def enable_auto_online(self):
        self.online.set(on=False)

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    voltage_doubler_charge_pump_power: F.ElectricPower
    """Terminal of the voltage doubler charge-pump capacitor """
    inverting_charge_pump_power: F.ElectricPower
    """Terminal of the inverting charge-pump capacitor """
    positive_charge_pump_power: F.ElectricPower
    """Regulated +5.5V output generated by the charge pump """
    negative_charge_pump_power: F.ElectricPower
    """Regulated -5.5V output generated by the charge pump """
    power: F.ElectricPower
    """Power input to the module"""

    uart: F.UART
    rs232: F.RS232

    enable: F.ElectricLogic
    online: F.ElectricLogic
    status: F.ElectricLogic

    cts_inverted: F.ElectricLogic

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://assets.maxlinear.com/web/documents/sp3243e.pdf"
    )

    @L.rt_field
    def descriptive_properties(self):
        return F.has_explicit_part.by_mfr("MaxLinear", "SP3243EBEA-L/TR")

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.inverting_charge_pump_power.hv: ["C2+"],
                self.rs232.rts.line: ["T2OUT"],
                self.rs232.dtr.line: ["T3OUT"],
                self.uart.dtr.line: ["T3IN"],
                self.uart.rts.line: ["T2IN"],
                self.uart.base_uart.tx.line: ["T1IN"],
                self.uart.ri.line: ["R5OUT"],
                self.uart.dcd.line: ["R4OUT"],
                self.uart.dsr.line: ["R3OUT"],
                self.uart.cts.line: ["R2OUT"],
                self.uart.base_uart.rx.line: ["R1OUT"],
                self.inverting_charge_pump_power.lv: ["C2-"],
                self.cts_inverted.line: ["R2OUT#"],
                self.status.line: ["STATUS#"],
                self.enable.line: ["SHUTDOWN#"],
                self.online.line: ["ONLINE#"],
                self.voltage_doubler_charge_pump_power.lv: ["C1-"],
                self.power.lv: ["GND"],
                self.power.hv: ["VCC"],
                self.positive_charge_pump_power.hv: ["V+"],
                self.voltage_doubler_charge_pump_power.hv: ["C1+"],
                self.negative_charge_pump_power.lv: ["V-"],
                self.rs232.rx.line: ["R1IN"],
                self.rs232.cts.line: ["R2IN"],
                self.rs232.dsr.line: ["R3IN"],
                self.rs232.dcd.line: ["R4IN"],
                self.rs232.ri.line: ["R5IN"],
                self.rs232.tx.line: ["T1OUT"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.power.voltage.constrain_subset(L.Range(3.0 * P.V, 5.5 * P.V))

        self.uart.base_uart.baud.constrain_le(250 * P.kbaud)

        self.rs232.get_trait(
            F.has_single_electric_reference
        ).get_reference().voltage.constrain_subset(
            L.Range.from_center(3 * P.V, 15 * P.V)
        )  # TODO: Support negative numbers (-15 * P.V, 15 * P.V))

        # ------------------------------------
        #           connections
        # ------------------------------------
        F.ElectricLogic.connect_all_module_references(self, exclude=[self.rs232])

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.power.voltage.constrain_subset(L.Range(3.0 * P.V, 5.5 * P.V))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import math
from operator import add
from typing import Iterable, Sequence, TypeVar

import numpy as np
from shapely import MultiPolygon, Point, Polygon, transform

logger = logging.getLogger(__name__)


def polygon_insert_cutout(polygon: Polygon, cutout: Polygon) -> Polygon:
    """
    Insert a cutout into a polygon

    :param polygon: The polygon to insert the cutout into
    :param cutout: The cutout to insert
    :return: The polygon with the cutout inserted
    """

    p_coords = list(polygon.exterior.coords)
    c_coords = list(cutout.exterior.coords)

    dist = [
        (Geometry.distance_euclid(p1, p2), p1, p2) for p1, p2 in zip(p_coords, c_coords)
    ]

    _, p_closest, c_closest = min(dist, key=lambda x: x[0])

    # roll the cutout so that the closest point is first, then append the first point to
    # close the cutout
    c_first_idx = c_coords.index(c_closest)
    c_coords = list(np.roll(c_coords, -c_first_idx, axis=0))
    c_coords.append(c_coords[0])

    # Insert the cutout into the polygon
    p_first_idx = p_coords.index(p_closest)
    flattened_coords = (
        p_coords[:p_first_idx]
        + [p_coords[p_first_idx]]
        + c_coords
        + p_coords[p_first_idx:]
    )

    return Polygon(flattened_coords)


def flatten_polygons(polygons: list[Polygon]) -> list[Polygon]:
    """
    Flatten a list of polygons by removing polygons that are contained in other polygons

    :param polygons: The polygons to flatten
    :return: The flattened polygons
    """
    flattened_polygons = []
    while polygons:
        remaining = []
        for poly in polygons:
            # skip lower level polygons
            if any([p.contains(poly) for p in polygons if p != poly]):
                remaining.append(poly)
                continue
            # cut out polygons that are contained in other polygons
            if any([z.contains(poly) for z in flattened_polygons]):
                flattened_polygons = [
                    polygon_insert_cutout(z, poly) if z.contains(poly) else z
                    for z in flattened_polygons
                ]
            else:
                # add polygons that are not contained in any other polygon
                flattened_polygons.append(poly)
        polygons = remaining

    return flattened_polygons


def get_distributed_points_in_polygon(
    polygon: Polygon,
    density: float,
) -> list[Point]:
    """
    Get a list of points that are distributed in a polygon

    :param polygon: The polygon to distribute the points in
    :param density: The density of the points
    :return: A list of points that are distributed in the polygon
    """

    num_points = int(polygon.area * density)
    if polygon.area > 0 and num_points == 0:
        num_points = 1

    points = get_random_points_in_polygon(polygon, num_points)

    min_x, min_y, max_x, max_y = polygon.bounds
    size_x = max_x - min_x
    size_y = max_y - min_y

    while True:
        point_distance_travel = []
        for i, point in enumerate(points):
            point_bubble = polygon

            for p in points:
                if p == point:
                    continue

                p1 = np.array((point.x, point.y))
                p2 = np.array((p.x, p.y))

                p_mid = (p1 + p2) / 2

                v = p2 - p1
                v_unit = v / np.linalg.norm(v)
                v_unit_perp = np.array([v_unit[1], -v_unit[0]])

                # create a polygon at least twice as large to cut the polygon up
                exclusion_polygon = Polygon(
                    [
                        (
                            p_mid[0] + v_unit_perp[0] * size_x / 2,
                            p_mid[1] + v_unit_perp[1] * size_x / 2,
                        ),
                        (
                            p_mid[0] + v_unit_perp[0] * size_x / 2 + v_unit[0] * size_y,
                            p_mid[1] + v_unit_perp[1] * size_x / 2 + v_unit[1] * size_y,
                        ),
                        (
                            p_mid[0] - v_unit_perp[0] * size_x / 2 + v_unit[0] * size_y,
                            p_mid[1] - v_unit_perp[1] * size_x / 2 + v_unit[1] * size_y,
                        ),
                        (
                            p_mid[0] - v_unit_perp[0] * size_x,
                            p_mid[1] - v_unit_perp[1] * size_x,
                        ),
                    ]
                )

                point_bubble = point_bubble.difference(exclusion_polygon)

            assert point_bubble.contains(point)

            if isinstance(point_bubble, MultiPolygon):
                for p in list(point_bubble.geoms):
                    if p.contains(point):
                        point_bubble = p
                        break

            new_point = point_bubble.centroid

            # TODO: it is possible it will be outside, then just move it to the
            # closest point on the polygon.
            if not point_bubble.contains(new_point):
                continue

            assert polygon.contains(new_point)

            point_distance_travel.append(
                Geometry.distance_euclid(
                    Geometry.Point((point.x, point.y)),
                    Geometry.Point((new_point.x, new_point.y)),
                )
            )
            logger.debug(
                f"Point ({point.x:.2f}, {point.y:.2f}) -> ({new_point.x:.2f}, "
                f"{new_point.y:.2f}), distance: {point_distance_travel[-1]}"
            )

            points[i] = new_point

        if max(point_distance_travel) < density / 100:
            break

    return points


def closest_point_on_segment_to_point(line: tuple[Point, Point], point: Point) -> Point:
    """
    Finds the minimum distance from the point to the line defined by two points

    :param line: The line segment defined by two points
    :param point: The point to find the distance to
    :return: The closest point on the line segment to the point
    """
    if line[0] == line[1]:
        return line[0]

    p1 = np.array((line[0].x, line[0].y))
    p2 = np.array((line[1].x, line[1].y))
    p3 = np.array((point.x, point.y))

    v_line = p2 - p1
    v_point = p3 - p1

    # Project point_vec onto line_vec
    v_line_unit = v_line / np.linalg.norm(v_line)

    # Unit vector in direction of line_vec
    proj_length = np.dot(v_point, v_line_unit)  # Projection length

    if proj_length < 0:
        # The projection falls before p1
        closest_point = p1
    elif proj_length > np.linalg.norm(v_line):
        # The projection falls after p2
        closest_point = p2
    else:
        # The projection falls within the segment
        proj_vec = proj_length * v_line_unit  # Projection vector
        closest_point = p1 + proj_vec

    p = Point(*closest_point)

    assert p != Point(), f"Could not find closes point on line {line} to point {point}"

    return p


def closest_point_on_line(p1, p2, p3):
    """
    Finds the closest point on the line defined by p1 and p2 to the point p3.

    Parameters:
    p1, p2, p3: Arrays or lists representing the coordinates of the points.

    Returns:
    The coordinates of the closest point on the line to p3.
    """
    # Convert points to numpy arrays
    p1 = np.array(p1)
    p2 = np.array(p2)
    p3 = np.array(p3)

    # Vector from p1 to p2
    line_vec = p2 - p1
    # Vector from p1 to p3
    point_vec = p3 - p1

    # Project point_vec onto line_vec
    line_vec_unit = line_vec / np.linalg.norm(
        line_vec
    )  # Unit vector in direction of line_vec
    proj_length = np.dot(point_vec, line_vec_unit)  # Projection length
    proj_vec = proj_length * line_vec_unit  # Projection vector

    # Closest point on the line
    closest_point = p1 + proj_vec

    return closest_point


def get_random_points_in_polygon(polygon: Polygon, num_points: int) -> list[Point]:
    """
    Get a list of unique random points that are inside a polygon

    :param polygon: The polygon to get the points from
    :param num_points: The number of points to get
    :return: A list of unique random points that are inside the polygon
    """
    points = []
    min_x, min_y, max_x, max_y = polygon.bounds
    while len(points) < num_points:
        p = Point(np.random.uniform(min_x, max_x), np.random.uniform(min_y, max_y))
        if polygon.contains(p) and p not in points:
            points.append(p)
    return points


def polygon_to_lines(polygon: Polygon) -> Iterable[tuple[Point, Point]]:
    coords = [Point(p) for p in polygon.exterior.coords]
    coords += [coords[0]]
    for i in range(len(polygon.exterior.coords) - 1):
        yield (coords[i], coords[i + 1])


def intersect_polygon_with_grid(
    polys: list[Polygon],
    grid_pitch: tuple[float, float],
    grid_offset: tuple[float, float],
) -> list[Point]:
    """
    Get a list of points on a grid that are inside a list of non-intersecting polygons

    :param polys: The polygons to check, can be either an exterior or interior polygon
    :param grid_pitch: The pitch of the grid (x, y)
    :param grid_offset: The offset of the grid (x, y)
    :return: A list of points that are inside the polygons
    """

    points = []
    min_x, min_y, max_x, max_y = (math.inf, math.inf, -math.inf, -math.inf)
    for b in [(p.bounds) for p in polys]:
        min_x = min(min_x, b[0])
        min_y = min(min_y, b[1])
        max_x = max(max_x, b[2])
        max_y = max(max_y, b[3])

    grid_start_x = math.ceil(min_x / grid_pitch[0]) * grid_pitch[0] + grid_offset[0]
    grid_start_y = math.ceil(min_y / grid_pitch[1]) * grid_pitch[1] + grid_offset[1]
    grid_x = np.arange(grid_start_x, max_x, grid_pitch[0])
    grid_y = np.arange(grid_start_y, max_y, grid_pitch[1])

    fill_polys = []
    keepout_polys = []

    # only support one level of keepout, no fills inside keepouts
    for poly in polys:
        if any([p.contains(poly) for p in polys if p != poly]):
            keepout_polys.append(poly)
            continue
        fill_polys.append(poly)

    for poly in fill_polys:
        points += [
            Point(x, y)
            for x in grid_x
            for y in grid_y
            if poly.contains(Point(x, y))
            and not any([p.contains(Point(x, y)) for p in keepout_polys])
        ]

    return points


def transform_polygon(
    poly: Polygon, offset: tuple[float, float], scale: tuple[float, float]
) -> Polygon:
    """
    Transform a list of polygons using a transformation matrix

    :param polys: The polygons to transform
    :param offset: The offset to apply (x, y)
    :param scale: The scale to apply (x, y)

    :return: The transformed polygons
    """

    return transform(poly, lambda x: (x + offset) * scale)


# TODO: cleanup and merge
class Geometry:
    Point2D = tuple[float, float]
    Point = tuple[float, float, float, int]

    @staticmethod
    def add_points(p1: Point | Point2D, p2: Point | Point2D) -> Point | Point2D:
        p1e = p1 + (0,) * (4 - len(p1))
        p2e = p2 + (0,) * (4 - len(p2))
        out = tuple(map(add, p1e, p2e))
        if len(p1) == len(p2) == 2:
            return out[:2]
        return out

    # class Point2D(tuple[float, float]):
    #    def __add__(self, other: "Geometry.Point2D") -> "Geometry.Point2D":
    #        return Geometry.Point2D(self[0] + other[0], self[1] + other[1])

    ## TODO fix all Point2D functions to use Point

    ## TODO more generic
    ## x,y, rotation, layer
    ## Point = tuple[float, float, float, int]

    # class Point(tuple[float, float, float, int]):
    #    def __round__(self, ndigits=None) -> Point:
    #        return Point(
    #            round(self[0], ndigits=ndigits),
    #            round(self[1], ndigits=ndigits),
    #            round(self[2], ndigits=ndigits),
    #            self[3],
    #        )

    #    def twod(self) -> "Geometry.Point2D":
    #        return Geometry.Point2D(self[:2])

    #    @property
    #    def x(self) -> float:
    #        return self[0]

    #    @property
    #    def y(self) -> float:
    #        return self[1]

    #    @property
    #    def rotation_deg(self) -> float:
    #        return self[2]

    #    @property
    #    def layer(self) -> int:
    #        return self[3]

    #    def __add__(self, other: "Point | Geometry.Point2D") -> Point:
    #        if isinstance(other, Geometry.Point2D):
    #            other = Point(other[0], other[1], 0, self.layer)

    #        assert self.layer == other.layer

    #        return Point(
    #            self.x + other.x,
    #            self.y + other.y,
    #            # TODO does this make sense?
    #            self.rotation_deg + other.rotation_deg,
    #            self.layer,
    #        )

    @staticmethod
    def mirror(axis: tuple[float | None, float | None], structure: list[Point2D]):
        return [
            (
                2 * axis[0] - x if axis[0] is not None else x,
                2 * axis[1] - y if axis[1] is not None else y,
            )
            for (x, y) in structure
        ]

    @staticmethod
    def as2d(coord: Point | Point2D) -> Point2D:
        return coord[:2]

    @staticmethod
    def as4d(coord: Point | Point2D) -> Point:
        return coord + (0,) * (4 - len(coord))

    @staticmethod
    def abs_pos(parent_: Point | Point2D, child_: Point | Point2D) -> Point:
        # Extend to x,y,rotation,layer
        parent = Geometry.as4d(parent_)
        child = Geometry.as4d(child_)

        rot_parent = parent[2]
        rot_child = child[2]
        total_rot_deg = rot_parent + rot_child

        # Rotate child vector around parent and add
        # rotation deg is negative because kicad rotates counter clock wise
        parent_absolute_xy = parent[:2]
        child_relative_xy = child[:2]
        x, y = parent_absolute_xy
        cx_rotated, cy_rotated = Geometry.rotate(
            (0, 0), [child_relative_xy], -rot_parent
        )[0]
        abs_x = x + cx_rotated
        abs_y = y + cy_rotated

        # Layer
        parent_layer = parent[3]
        child_layer = child[3]
        if parent_layer != 0 and child_layer != 0:
            raise Exception(
                f"Adding two non-zero layers: {parent_layer=} + {child_layer=}"
            )
        layer = parent_layer + child_layer

        out = (
            abs_x,
            abs_y,
            total_rot_deg,
            layer,
        )

        return out

    @staticmethod
    def translate(vec: Point2D, structure: list[Point2D]) -> list[Point2D]:
        return [tuple(map(add, vec, point)) for point in structure]

    @classmethod
    def rotate(
        cls, axis: Point2D, structure: list[Point2D], angle_deg: float
    ) -> list[Point2D]:
        theta = np.radians(angle_deg)
        c, s = np.cos(theta), np.sin(theta)
        R = np.array(((c, -s), (s, c)))

        return cls.translate(
            axis,
            [
                tuple(R @ np.array(point))
                for point in cls.translate((-axis[0], -axis[1]), structure)
            ],
        )

    C = TypeVar("C", tuple[float, float], tuple[float, float, float])

    @staticmethod
    def triangle(start: C, width: float, depth: float, count: int):
        x1, y1 = start[:2]

        n = count - 1
        cy = width / n

        ys = [round(y1 + cy * i, 2) for i in range(count)]
        xs = [round(x1 + depth * (1 - abs(1 - 1 / n * i * 2)), 2) for i in range(count)]

        return list(zip(xs, ys))

    @staticmethod
    def line(start: C, length: float, count: int):
        x1, y1 = start[:2]

        n = count - 1
        cy = length / n

        ys = [round(y1 + cy * i, 2) for i in range(count)]
        xs = [x1] * count

        return list(zip(xs, ys))

    @staticmethod
    def line2(start: C, end: C, count: int):
        x1, y1 = start[:2]
        x2, y2 = end[:2]

        n = count - 1
        cx = (x2 - x1) / n
        cy = (y2 - y1) / n

        ys = [round(y1 + cy * i, 2) for i in range(count)]
        xs = [round(x1 + cx * i, 2) for i in range(count)]

        return list(zip(xs, ys))

    @staticmethod
    def find_circle_center(p1, p2, p3):
        """
        Finds the center of the circle passing through the three given points.
        """
        # Compute the midpoints
        mid1 = (p1 + p2) / 2
        mid2 = (p2 + p3) / 2

        # Compute the slopes of the lines
        m1 = (p2[1] - p1[1]) / (p2[0] - p1[0])
        m2 = (p3[1] - p2[1]) / (p3[0] - p2[0])

        # The slopes of the perpendicular bisectors
        perp_m1 = -1 / m1
        perp_m2 = -1 / m2

        # Equations of the perpendicular bisectors
        # y = perp_m1 * (x - mid1[0]) + mid1[1]
        # y = perp_m2 * (x - mid2[0]) + mid2[1]

        # Solving for x
        x = (mid2[1] - mid1[1] + perp_m1 * mid1[0] - perp_m2 * mid2[0]) / (
            perp_m1 - perp_m2
        )

        # Solving for y using one of the bisector equations
        y = perp_m1 * (x - mid1[0]) + mid1[1]

        return np.array([x, y])

    @staticmethod
    def approximate_arc(
        p_start: Point2D, p_mid: Point2D, p_end: Point2D, resolution=10
    ) -> list[tuple[Point2D, Point2D]]:
        p_start, p_mid, p_end = (np.array(p) for p in (p_start, p_mid, p_end))

        # Calculate the center of the circle
        center = Geometry.find_circle_center(p_start, p_mid, p_end)

        # Calculate start, mid, and end angles
        start_angle = np.arctan2(p_start[1] - center[1], p_start[0] - center[0])
        mid_angle = np.arctan2(p_mid[1] - center[1], p_mid[0] - center[0])
        end_angle = np.arctan2(p_end[1] - center[1], p_end[0] - center[0])

        # Adjust angles if necessary
        if start_angle > mid_angle:
            start_angle -= 2 * np.pi
        if mid_angle > end_angle:
            mid_angle -= 2 * np.pi

        # Radius of the circle
        r = np.linalg.norm(p_start - center)

        # Compute angles of line segment endpoints
        angles = np.linspace(start_angle, end_angle, resolution + 1)

        # Compute points on the arc
        points = np.array(
            [[center[0] + r * np.cos(a), center[1] + r * np.sin(a)] for a in angles]
        )

        # Create line segments
        segments = [(points[i], points[i + 1]) for i in range(resolution)]

        seg_no_np = [(tuple(a), tuple(b)) for a, b in segments]

        return seg_no_np

    @staticmethod
    def distance_euclid(p1: Point, p2: Point):
        return math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)

    @staticmethod
    def bbox(
        points: Sequence[Point | Point2D], tolerance=0.0
    ) -> tuple[Point2D, Point2D]:
        min_x = min(p[0] for p in points)
        min_y = min(p[1] for p in points)
        max_x = max(p[0] for p in points)
        max_y = max(p[1] for p in points)

        return (
            min_x - tolerance,
            min_y - tolerance,
        ), (
            max_x + tolerance,
            max_y + tolerance,
        )

    @staticmethod
    def rect_to_polygon(rect: tuple[Point2D, Point2D]) -> list[Point2D]:
        return [
            (rect[0][0], rect[0][1]),
            (rect[1][0], rect[0][1]),
            (rect[1][0], rect[1][1]),
            (rect[0][0], rect[1][1]),
        ]

    @staticmethod
    def average(points: list[Point2D | Point]) -> Point:
        points4d = [p + (0,) * (4 - len(p)) for p in points]
        # same layer
        assert len({p[3] for p in points4d}) == 1
        points2d = [p[:2] for p in points4d]

        out = tuple(np.mean(points2d, axis=0))
        if any(len(p) > 2 for p in points):
            return out + (0, points[0][3])



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from more_itertools import first

import faebryk.library._F as F
from faebryk.libs.util import not_none


class has_footprint_impl(F.has_footprint.impl()):
    def set_footprint(self, fp: F.Footprint):
        self.obj.add(fp, name="footprint")

    def try_get_footprint(self) -> F.Footprint | None:
        if fps := self.obj.get_children(direct_only=True, types=F.Footprint):
            assert len(fps) == 1, f"In obj: {self.obj}: candidates: {fps}"
            return first(fps)
        else:
            return None

    def get_footprint(self) -> F.Footprint:
        return not_none(self.try_get_footprint())



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging
from enum import Enum

from more_itertools import first

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.libs.library import L
from faebryk.libs.sets.sets import P_Set
from faebryk.libs.units import P
from faebryk.libs.util import invert_dict, md_list, partition_as_list

logger = logging.getLogger(__name__)


class I2C(ModuleInterface):
    scl: F.ElectricLogic
    sda: F.ElectricLogic

    address = L.p_field(
        within=L.Range(0, 0x7F),
        domain=L.Domains.Numbers.NATURAL(),
    )
    bus_addresses = L.p_field(
        within=L.Range(0, 0x7F),
        domain=L.Domains.Numbers.NATURAL(),
    )

    frequency = L.p_field(
        units=P.Hz,
        likely_constrained=True,
        soft_set=L.Range(10 * P.kHz, 3.4 * P.MHz),
    )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    @L.rt_field
    def requires_pulls(self):
        def pred(signal: F.ElectricSignal, bus: set[Node]):
            interface = signal.get_parent_of_type(I2C)

            assert interface in bus

            return (
                len(bus) > 1
                # arbitrarily choose an interface to represent the bus for this check
                and first(sorted(bus, key=lambda n: str(n))) is interface
                # indicates usage
                and signal.line.net_crosses_pad_boundary()
            )

        return F.requires_pulls(
            self.scl,
            self.sda,
            pred=pred,
            required_resistance=L.Range(
                1 * (1 - 0.1) * P.kohm, 10 * (1 + 0.1) * P.kohm
            ),
        )

    def bus_crosses_pad_boundary(self):
        return (
            self.scl.line.net_crosses_pad_boundary()
            or self.sda.line.net_crosses_pad_boundary()
        )

    def terminate(self, owner: Module):
        # TODO: https://www.ti.com/lit/an/slva689/slva689.pdf

        self.pull_up_sda = self.sda.pulled.pull(up=True, owner=owner)
        self.pull_up_scl = self.scl.pulled.pull(up=True, owner=owner)

    class SpeedMode(Enum):
        low_speed = 10 * P.khertz
        standard_speed = 100 * P.khertz
        fast_speed = 400 * P.khertz
        high_speed = 3.4 * P.Mhertz

    @staticmethod
    def define_max_frequency_capability(mode: SpeedMode):
        return L.Range(I2C.SpeedMode.low_speed.value, mode.value)

    def __preinit__(self) -> None:
        self.frequency.add(F.is_bus_parameter())
        # self.bus_addresses.add(F.is_bus_parameter(reduce=(self.address, Union)))

    def _hack_get_connected(self):
        """
        Workaround for hierarchical mifs not working currently
        """
        # Find all I2C interfaces connected to the same bus (via SCL line)
        # Assumption: If SCL is connected, SDA is also connected to the same
        # set of interfaces
        # Ensure the signal is connected to a line

        # Get all nodes connected electrically to the line
        connected_nodes = self.scl.line.get_connected()
        # Get all nodes connected logically to the line
        connected_nodes |= self.sda.get_connected()

        bus_interfaces: set[I2C] = set()
        for node in connected_nodes:
            interface = node.get_parent_of_type(I2C)
            # Filter out nodes not part of an I2C interface
            if interface is not None:
                bus_interfaces.add(interface)

        # include shallow connections
        bus_interfaces |= self.get_connected(include_self=False).keys()

        return bus_interfaces

    class requires_unique_addresses(ModuleInterface.TraitT.decless()):
        class DuplicateAddressException(
            F.implements_design_check.UnfulfilledCheckException
        ):
            def __init__(self, duplicates: dict[P_Set, list["I2C"]], bus: set["I2C"]):
                message = "Duplicate I2C addresses found on the bus:\\n"
                message += md_list(
                    {f"{addr}": nodes for addr, nodes in duplicates.items()}
                )
                super().__init__(message, nodes=list(bus))

        design_check: F.implements_design_check

        @F.implements_design_check.register_post_solve_check
        def __check_post_solve__(self):
            solver = self.design_check.get_solver()
            obj = self.get_obj(I2C)
            bus_interfaces = obj._hack_get_connected()

            # If only self or less found, no conflicts possible
            if len(bus_interfaces) <= 1:
                return

            # Get addresses, handling potential unresolved parameters gracefully
            addresses: dict[I2C, P_Set] = {
                interface: solver.inspect_get_known_supersets(interface.address)
                for interface in bus_interfaces
            }
            unresolved, resolved = partition_as_list(
                lambda s: s[1].is_single_element(), addresses.items()
            )

            # Check for duplicates
            by_id = invert_dict(dict(resolved))
            duplicates = {
                addr: busses for addr, busses in by_id.items() if len(busses) > 1
            }
            if duplicates:
                raise I2C.requires_unique_addresses.DuplicateAddressException(
                    duplicates=duplicates, bus=bus_interfaces
                )

            # TODO: Consider raising MaybeUnfulfilled if there are unresolved addresses?
            # For now, we only raise if we find concrete duplicates.

    address_check: requires_unique_addresses



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

import faebryk.library._F as F
from faebryk.core.graph import GraphFunctions
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.solver.solver import Solver
from faebryk.libs.util import DAG, groupby, partition_as_list

logger = logging.getLogger(__name__)


def export_i2c_tree(
    app: Module, solver: Solver, path: Path = Path("build/documentation/i2c_tree.md")
):
    """
    Export the I2C tree of the given application to a file.
    """

    # Filter buses
    mifs = GraphFunctions(app.get_graph()).nodes_of_type(F.I2C)
    buses = ModuleInterface._group_into_buses(mifs)
    buses = {
        k: v for k, v in buses.items() if len(v) > 1 and k.bus_crosses_pad_boundary()
    }
    buses_with_address = [
        {
            mif: lit
            for mif in bus
            if (
                lit := solver.inspect_get_known_supersets(mif.address)
            ).is_single_element()
        }
        for bus in buses.values()
    ]

    # Build DAG
    dag = DAG[F.I2C]()
    for bus in buses_with_address:
        clients, servers = partition_as_list(lambda x: x[1] == 0, bus.items())
        for client in clients:
            client_mif, _ = client
            for server in servers:
                server_mif, _ = server
                dag.add_edge(server_mif, client_mif)

    grouped_by_parent = groupby(
        dag.values, lambda x: p[0] if (p := x.get_parent()) is not None else None
    )

    # Make a mermaid diagram from the dag
    out = "graph TD\n"
    node_names = {}
    i = 0
    # Build nodes
    for container, mifs in grouped_by_parent.items():
        if len(mifs) > 1 and container is not None:
            out += f"    subgraph {container.get_full_name()}\n"
        for mif in mifs:
            node_name = f"node_{i}"
            node_names[mif] = node_name
            address = solver.inspect_get_known_supersets(mif.address)

            label = f"{mif.get_full_name()}"
            is_server = address == 0

            if is_server:
                out += f'    {node_name}("{label}")\n'
            else:
                address_str = hex(int(address.any().m))
                label += f"<br/>{address_str}"
                out += f'    {node_name}(("{label}"))\n'

            i += 1

        if len(mifs) > 1 and container is not None:
            out += "    end\n"

    # Build edges
    for parent_value in dag.values:
        parent_internal_node = dag.get(parent_value)
        parent_node_name = node_names[parent_value]
        for child_internal_node in parent_internal_node._children:
            child_value = child_internal_node.value
            child_node_name = node_names[child_value]
            out += f"    {parent_node_name} --> {child_node_name}\n"

    # Write to file
    path.parent.mkdir(parents=True, exist_ok=True)
    markdown = "```mermaid\n" + out + "```"
    path.write_text(markdown, encoding="utf-8")
    logger.info(f"Wrote I2C tree to {path}")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from deprecated import deprecated

import faebryk.library._F as F

Resistor_Voltage_Divider = deprecated(
    reason="Use ResistorVoltageDivider instead",
)(F.ResistorVoltageDivider)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Potentiometer(Module):
    resistors_ifs = L.list_field(2, F.Electrical)
    wiper: F.Electrical
    total_resistance = L.p_field(units=P.ohm)
    resistors = L.list_field(2, F.Resistor)

    def __preinit__(self):
        for i, resistor in enumerate(self.resistors):
            self.resistors_ifs[i].connect_via(resistor, self.wiper)

        self.total_resistance.alias_is(
            self.resistors[0].resistance + self.resistors[1].resistance
        )

    def connect_as_voltage_divider(
        self, high: F.Electrical, low: F.Electrical, out: F.Electrical
    ):
        self.resistors_ifs[0].connect(high)
        self.resistors_ifs[1].connect(low)
        self.wiper.connect(out)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class GDT(Module):
    common: F.Electrical
    tube_1: F.Electrical
    tube_2: F.Electrical

    dc_breakdown_voltage = L.p_field(
        units=P.V,
        likely_constrained=True,
        soft_set=L.Range(100 * P.V, 1000 * P.V),
    )
    impulse_discharge_current = L.p_field(
        units=P.A,
        likely_constrained=True,
        soft_set=L.Range(100 * P.mA, 100 * P.A),
    )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.tube_1, self.tube_2)

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.GDT
    )



from faebryk.core.node import Node
from faebryk.core.reference import Reference
from faebryk.core.trait import Trait


class has_reference(Trait):
    """Trait-attached reference"""

    reference = Reference()

    def __init__(self, reference: Node):
        super().__init__()
        self.reference = reference



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import unittest
from typing import Sequence

from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.libs.library import L


class TestUtil(unittest.TestCase):
    def test_trees(self):
        class N(Module):
            mif: ModuleInterface

            @L.rt_field
            def n(self):
                if self._depth == 0:
                    # TODO does this work?
                    return []
                return N(self._depth - 1)

            def __init__(self, depth: int):
                super().__init__()
                self._depth = depth

        level_count = 5
        n = N(level_count)

        def assertEqual(n1: Sequence[Node], n2: Sequence[Node]):
            n1s = list(sorted(n1, key=id))
            n2s = list(sorted(n2, key=id))
            self.assertEqual(n1s, n2s)

        tree = n.get_tree(types=(Module, ModuleInterface))
        levels = list(tree.iter_by_depth())
        print(tree)
        for i, le in enumerate(levels):
            print(i, le)
        self.assertEqual(len(levels), level_count + 2)
        assertEqual(levels[0], [n])
        n_i = n
        for i in range(1, level_count + 1):
            assertEqual(levels[i], [n_i.n, n_i.mif])
            n_i = n_i.n
        assertEqual(levels[level_count + 1], [n_i.mif])



import os
import shutil
import sys
from pathlib import Path

import pytest

from faebryk.libs.util import repo_root as _repo_root
from faebryk.libs.util import run_live

# Get the examples directory relative to this test file
EXAMPLES_DIR = _repo_root() / "examples"


@pytest.mark.xfail(reason="Need to update atopile/packages to use atomic parts")
@pytest.mark.parametrize("package", ["atopile/addressable-leds"])
def test_install_package(package: str, tmp_path: Path):
    example_copy = tmp_path / "example"

    shutil.copytree(
        EXAMPLES_DIR,
        example_copy,
        ignore=lambda src, names: [".ato", "build", "standalone"],
    )

    _, stderr, _ = run_live(
        [sys.executable, "-m", "atopile", "add", package],
        env={**os.environ, "NONINTERACTIVE": "1"},
        cwd=example_copy,
        stdout=print,
        stderr=print,
    )

    assert f"Installing {package} to" in stderr
    assert "Done adding" in stderr.splitlines()[-1]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module


class has_single_electric_reference_shared(F.has_single_electric_reference.impl()):
    lazy: F.is_lazy

    def __init__(self, gnd_only: bool = False):
        super().__init__()
        self.gnd_only = gnd_only

    def on_obj_set(self):
        super().on_obj_set()

        obj = self.get_obj(Module)

        F.ElectricSignal.connect_all_module_references(obj, gnd_only=self.gnd_only)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class DE9Connector(Module):
    """
    DE-9 connector
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    contact = L.list_field(9, F.Electrical)
    shield: F.Electrical

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.X
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.shield: ["MH1"],
                self.shield: ["MH2"],
                self.contact[0]: ["1"],
                self.contact[1]: ["2"],
                self.contact[2]: ["3"],
                self.contact[3]: ["4"],
                self.contact[4]: ["5"],
                self.contact[5]: ["6"],
                self.contact[6]: ["7"],
                self.contact[7]: ["8"],
                self.contact[8]: ["9"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    @L.rt_field
    def can_attach_to_footprint(self):
        pinmap = {f"{i + 1}": ei for i, ei in enumerate(self.contact)}
        pinmap.update({"10": self.shield})
        return F.can_attach_to_footprint_via_pinmap(pinmap)

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface


class USB3_IF(ModuleInterface):
    usb_if: F.USB2_0_IF
    rx: F.DifferentialPair
    tx: F.DifferentialPair
    gnd_drain: F.Electrical



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L


class USB_C_PowerOnly(ModuleInterface):
    power: F.ElectricPower
    cc1: F.Electrical
    cc2: F.Electrical

    @L.rt_field
    def surge_protected(self):
        return F.can_be_surge_protected_defined(
            self.power.lv,
            self.power.hv,
            self.cc1,
            self.cc2,
        )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def connect_to_full_usb_c(self, usb_c: F.USB_C):
        self.power.connect(usb_c.usb3.usb3_if.usb_if.buspower)
        self.cc1.connect(usb_c.cc1)
        self.cc2.connect(usb_c.cc2)
        return self



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import inspect
import logging
from typing import Callable, Iterable, cast

from more_itertools import first

import faebryk.library._F as F
from atopile import errors
from faebryk.core.cpp import Path
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.trait import Trait
from faebryk.libs.exceptions import accumulate
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class ERCFault(errors.UserException):
    """Base class for ERC faults."""


class ModuleInterfacePath(list[ModuleInterface]):
    """A path of ModuleInterfaces."""

    @classmethod
    def from_path(cls, path: Path) -> "ModuleInterfacePath":
        """
        Convert a Path (of GraphInterfaces) to a ModuleInterfacePath.
        """
        mifs = cast(
            list[ModuleInterface],
            [gif.node for gif in path if gif.node.isinstance(ModuleInterface)],
        )
        return cls(mifs)

    def snip_head(
        self, scissors: Callable[[ModuleInterface], bool], include_last: bool = True
    ) -> "ModuleInterfacePath":
        """
        Keep the head until the scissors predicate returns False.
        """
        for i, mif in enumerate(self):
            if not scissors(mif):
                return type(self)(self[: i + include_last])
        return self

    def snip_tail(
        self, scissors: Callable[[ModuleInterface], bool], include_first: bool = True
    ) -> "ModuleInterfacePath":
        """
        Keep the tail until the scissors predicate returns False.
        """
        for i, mif in reversed(list(enumerate(self))):
            if not scissors(mif):
                return type(self)(self[i + (not include_first) :])
        return self

    @classmethod
    def from_connection(
        cls, a: ModuleInterface, b: ModuleInterface
    ) -> "ModuleInterfacePath | None":
        """
        Return a ModuleInterfacePath between two ModuleInterfaces, if it exists,
        else None.
        """
        if paths := a.is_connected_to(b):
            # FIXME: Notes: from the master of graphs:
            #  - iterate through all paths
            #  - make a helper function
            #    ModuleInterfacePath.get_subpaths(path: Path, search: SubpathSearch)
            #    e.g SubpathSearch = tuple[Callable[[ModuleInterface], bool], ...]
            #  - choose out of subpaths
            #    - be careful with LinkDirectDerived edges (if there is a faulting edge
            #      is derived, save it as candidate and only yield it if no other found)
            #    - choose first shortest
            return cls.from_path(first(paths))
        return None


class ERCFaultShort(ERCFault):
    """Exception raised for short circuits."""


class ERCFaultShortedModuleInterfaces(ERCFaultShort):
    """Short circuit between two ModuleInterfaces."""

    def __init__(self, msg: str, path: ModuleInterfacePath, *args: object) -> None:
        super().__init__(msg, path, *args)
        self.path = path

    @classmethod
    def from_path(cls, path: ModuleInterfacePath) -> "ERCFaultShortedModuleInterfaces":
        """
        Given two shorted ModuleInterfaces, return an exception that describes the
        narrowest path for the fault.
        """
        return cls(f"`{' ~ '.join(mif.get_full_name() for mif in path)}`", path)


class ERCFaultElectricPowerUndefinedVoltage(ERCFault):
    def __init__(self, faulting_EP: F.ElectricPower, *args: object) -> None:
        msg = (
            f"ElectricPower with undefined or unsolved voltage: {faulting_EP}:"
            f" {faulting_EP.voltage}"
        )
        super().__init__(msg, [faulting_EP], *args)


class ERCPowerSourcesShortedError(ERCFault):
    """
    Multiple power sources shorted together
    """


def simple_erc(G: Graph, voltage_limit=1e5 * P.V):
    """Simple ERC check.

    This function will check for the following ERC violations:
    - shorted ElectricPower
    - shorted Caps
    - shorted Resistors
    - shorted symmetric footprints
    - shorted Nets
    - Net name collision

    - [unmapped pins for footprints]

    Args:

    Returns:
    """
    logger.info("Checking graph for ERC violations")

    with accumulate(ERCFault) as accumulator:
        # shorted power
        electricpower = GraphFunctions(G).nodes_of_type(F.ElectricPower)
        logger.info(f"Checking {len(electricpower)} Power")

        buses_grouped = ModuleInterface._group_into_buses(electricpower)
        buses = list(buses_grouped.values())

        # We do collection both inside and outside the loop because we don't
        # want to continue the loop if we've already raised a short exception
        with accumulator.collect():
            logger.info("Checking for hv/lv shorts")
            for ep in electricpower:
                if mif_path := ModuleInterfacePath.from_connection(ep.lv, ep.hv):

                    def _keep_head(x: ModuleInterface) -> bool:
                        if parent := x.get_parent():
                            parent_node, _ = parent
                            if isinstance(parent_node, F.ElectricPower):
                                return parent_node.hv is not x

                        return True

                    def _keep_tail(x: ModuleInterface) -> bool:
                        if parent := x.get_parent():
                            parent_node, _ = parent
                            if isinstance(parent_node, F.ElectricPower):
                                return parent_node.lv is not x

                        return True

                    raise ERCFaultShortedModuleInterfaces.from_path(
                        mif_path.snip_head(_keep_head).snip_tail(_keep_tail)
                    )

            logger.info("Checking for power source shorts")
            for bus in buses:
                with accumulator.collect():
                    sources = {
                        ep for ep in bus if ep.has_trait(F.Power.is_power_source)
                    }
                    if len(sources) <= 1:
                        continue

                    friendly_sources = ", ".join(n.get_full_name() for n in sources)
                    raise ERCPowerSourcesShortedError(
                        f"Power sources shorted: {friendly_sources}"
                    )

        # shorted nets
        nets = GraphFunctions(G).nodes_of_type(F.Net)
        logger.info(f"Checking {len(nets)} explicit nets")
        for net in nets:
            with accumulator.collect():
                collisions = {
                    p[0]
                    for mif in net.part_of.get_connected()
                    if (p := mif.get_parent()) and isinstance(p[0], F.Net)
                }

                if collisions:
                    friendly_shorted = ", ".join(n.get_full_name() for n in collisions)
                    raise ERCFaultShort(
                        f"Shorted nets: {friendly_shorted}",
                    )

        # shorted components
        # parts = [n for n in nodes if n.has_trait(has_footprint)]
        # sym_fps = [
        #    n.get_trait(has_footprint).get_footprint()
        #    for n in parts
        #    if n.has_trait(can_attach_to_footprint_symmetrically)
        # ]
        # logger.info(f"Checking {len(sym_fps)} symmetric footprints")
        # for fp in sym_fps:
        #    mifs = set(fp.get_all())
        #    checked = set()
        #    for mif in mifs:
        #        checked.add(mif)
        #        if any(mif.is_connected_to(other) for other in (mifs - checked)):
        #            raise ERCFault([mif], "shorted symmetric footprint")

        comps = GraphFunctions(G).nodes_of_types((F.Resistor, F.Capacitor, F.Fuse))
        logger.info(f"Checking {len(comps)} passives")
        for comp in comps:
            with accumulator.collect():
                assert isinstance(comp, (F.Resistor, F.Capacitor, F.Fuse))
                # TODO make prettier
                if (
                    comp.has_trait(F.has_part_picked)
                    and comp.get_trait(F.has_part_picked).removed
                ):
                    continue

                if path := ModuleInterfacePath.from_connection(
                    comp.unnamed[0], comp.unnamed[1]
                ):
                    raise ERCFaultShortedModuleInterfaces.from_path(path)

        ## unmapped Electricals
        # fps = [n for n in nodes if isinstance(n, Footprint)]
        # logger.info(f"Checking {len(fps)} footprints")
        # for fp in fps:
        #    for mif in fp.get_all():
        #        if not mif.get_direct_connections():
        #            raise ERCFault([mif], "no connections")

        # TODO check multiple pulls per logic


def check_modules_for_erc(module: Iterable[Module]):
    for m in module:
        logger.info(f"Checking {m} {'-' * 20}")
        simple_erc(m.get_graph())


def check_classes_for_erc(classes: Iterable[Callable[[], Module]]):
    modules = []
    for c in classes:
        try:
            m = c()
        except Exception as e:
            logger.warning(
                f"Could not instantiate {c.__name__}: {type(e).__name__}({e})"
            )
            continue
        modules.append(m)
    check_modules_for_erc(modules)


def check_library_for_erc(lib):
    members = inspect.getmembers(lib, inspect.isclass)
    module_classes = [m[1] for m in members if issubclass(m[1], Module)]
    check_classes_for_erc(module_classes)


# TODO split this up
class needs_erc_check(Trait.decless()):
    design_check: F.implements_design_check

    @F.implements_design_check.register_post_design_check
    def __check_post_design__(self):
        simple_erc(self.get_graph())



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging

import faebryk.library._F as F

logger = logging.getLogger(__name__)


class Addressor2(F.Addressor):
    """
    Curried Addressor for ato use
    """

    def __init__(self):
        super().__init__(address_bits=2)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass, field
from typing import Optional

import pytest
from dataclasses_json import CatchAll

import faebryk.library._F as F  # noqa: F401  # This is required to prevent a circular import
from faebryk.libs.kicad.fileformats_latest import C_kicad_pcb_file
from faebryk.libs.sexp.dataclass_sexp import (
    DecodeError,
    SymEnum,
    dataclass_dfs,
    dumps,
    loads,
    sexp_field,
)
from faebryk.libs.test.fileformats import (
    _FPLIB_DIR,  # noqa: F401
    _NETLIST_DIR,  # noqa: F401
    _PCB_DIR,  # noqa: F401
    _PRJ_DIR,  # noqa: F401
    _SCH_DIR,  # noqa: F401
    _SYM_DIR,  # noqa: F401
    _VERSION_DIR,  # noqa: F401
    DEFAULT_VERSION,  # noqa: F401
    PCBFILE,
)

logger = logging.getLogger(__name__)


def _unformat(s: str) -> str:
    return s.replace("\n", "").replace(" ", "")


@dataclass
class Container:
    @dataclass(kw_only=True)
    class SomeDataclass:
        class E_Numbers(SymEnum):
            ONE = "one"
            TWO = "two"

        a: int = field(**sexp_field(positional=True))
        b: bool = field(**sexp_field(positional=True))
        c: Optional[bool] = field(**sexp_field(positional=True), default=None)
        d: Optional[E_Numbers] = field(**sexp_field(positional=True), default=None)
        f: int
        g: bool
        h: Optional[bool] = None
        i: Optional[E_Numbers] = None

    some_dataclass: SomeDataclass


@pytest.mark.parametrize(
    "container,str_sexp",
    [
        (
            Container(
                some_dataclass=Container.SomeDataclass(
                    a=1, b=True, c=None, d=None, f=2, g=False, h=None, i=None
                )
            ),
            "(some_dataclass 1 yes (f 2) (g no))",
        ),
        (
            Container(
                some_dataclass=Container.SomeDataclass(
                    a=1,
                    b=True,
                    c=False,
                    d=Container.SomeDataclass.E_Numbers.ONE,
                    f=2,
                    g=False,
                    h=True,
                    i=Container.SomeDataclass.E_Numbers.TWO,
                )
            ),
            "(some_dataclass 1 yes no one (f 2) (g no) (h yes) (i two))",
        ),
    ],
)
def test_no_unknowns(container, str_sexp):
    cereal = dumps(container)

    assert _unformat(cereal) == _unformat(str_sexp)

    assert loads(cereal, Container) == container


def test_extra_positional():
    with pytest.raises(DecodeError):
        loads('(some_dataclass 1 yes no "one" gibberish (e 2) (f no))', Container)


def test_empty_unknowns():
    @dataclass
    class Container:
        @dataclass
        class SomeDataclass:
            a: int
            b: str
            unknown: CatchAll = None

        some_dataclass: SomeDataclass

    container = Container(some_dataclass=Container.SomeDataclass(a=1, b="hello"))

    cereal = dumps(container)

    assert _unformat(cereal) == _unformat('(some_dataclass (a 1) (b "hello"))')

    assert loads(cereal, Container) == container


def test_unknowns():
    @dataclass
    class Container:
        @dataclass
        class SomeDataclass:
            a: int
            b: str
            unknown: CatchAll = None

        some_dataclass: SomeDataclass

    cereal = (
        '(some_dataclass gibberish (thingo) (a 1) (b "hello") gibberish (thingo) '
        '(random_key "random_value") (whats_this (who_even_knows True)))'
    )

    container = loads(cereal, Container)

    assert container.some_dataclass.a == 1
    assert container.some_dataclass.b == "hello"
    assert container.some_dataclass.unknown  # there is content

    assert _unformat(dumps(container)) == _unformat(cereal)


def test_print_sexp():
    pcb = C_kicad_pcb_file.loads(PCBFILE)
    dfs = list(dataclass_dfs(pcb))
    for obj, path, name_path in dfs:
        name = "".join(name_path)
        logger.debug(f"{name:70} {[type(p).__name__ for p in path + [obj]]}")

    logger.debug("-" * 80)

    level2 = [p for p in dfs if len(p[1]) == 2]
    for obj, path, name_path in level2:
        name = "".join(name_path)
        logger.debug(f"{name:70}  {[type(p).__name__ for p in path + [obj]]}")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.exporters.pcb.layout.heuristic_decoupling import Params, place_next_to
from faebryk.exporters.pcb.layout.layout import Layout

logger = logging.getLogger(__name__)


class LayoutNextTo(Layout):
    Parameters = Params

    def __init__(self, target: F.Electrical, params: Params | None = None):
        super().__init__()
        self._params = params or Params()
        self._target = target

    def apply(self, *node: Node):
        # Remove nodes that have a position defined
        node = tuple(
            n
            for n in node
            if not n.has_trait(F.has_pcb_position) and n.has_trait(F.has_footprint)
        )

        for n in node:
            # TODO: why the assert?
            assert isinstance(n, Module)
            place_next_to(self._target, n, route=True, params=self._params)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging
from enum import Enum

logger = logging.getLogger(__name__)


class Iso262_MetricScrewThreadSizes(Enum):
    # TODO add tolerance etc
    # i think values are wrong, but api is ok
    M1 = 1.2
    M1_2 = 1.4
    M1_4 = 1.6
    M1_6 = 1.8
    M1_8 = 2.0
    M2 = 2.2
    M2_5 = 2.7
    M3 = 3.2
    M3_5 = 3.7
    M4 = 4.3
    M5 = 5.3
    M6 = 6.4
    M8 = 8.4
    M10 = 10.4
    M12 = 12.4
    M14 = 14.4
    M16 = 16.4
    M20 = 20.4
    M24 = 24.4
    M30 = 30.4
    M36 = 36.4
    M42 = 42.4
    M48 = 48.4
    M56 = 56.4
    M60 = 60.4
    M64 = 64.4



import pytest

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.app.checks import check_design
from faebryk.libs.exceptions import UserDesignCheckException
from faebryk.libs.library import L
from faebryk.libs.smd import SMDSize
from faebryk.libs.units import P


def test_i2c_requires_pulls():
    class A(Module):
        i2c: F.I2C

    class App(Module):
        a: A
        b: A

        def __preinit__(self):
            self.a.i2c.connect(self.b.i2c)

    app = App()

    # no issue if no pad boundary is crossed
    check_design(app.get_graph(), F.implements_design_check.CheckStage.POST_DESIGN)

    class App2(Module):
        a: A
        b: A

        def __preinit__(self):
            self.a.i2c.connect(self.b.i2c)

            self.a.add(
                F.can_attach_to_footprint_via_pinmap(
                    {
                        "1": self.a.i2c.sda.line,
                        "2": self.a.i2c.scl.line,
                    },
                )
            ).attach(F.SMDTwoPin(SMDSize.I0402, F.SMDTwoPin.Type.Resistor))
            self.b.add(
                F.can_attach_to_footprint_via_pinmap(
                    {
                        "1": self.b.i2c.sda.line,
                        "2": self.b.i2c.scl.line,
                    },
                )
            ).attach(F.SMDTwoPin(SMDSize.I0402, F.SMDTwoPin.Type.Resistor))

    app2 = App2()

    # required resistance can be customized
    app2.a.i2c.get_trait(F.requires_pulls).required_resistance = L.Range(
        0.1 * P.kohm, 0.5 * P.kohm
    )

    # connection crosses pad boundary, so the check now fails
    with pytest.raises(UserDesignCheckException):
        check_design(app2.get_graph(), F.implements_design_check.CheckStage.POST_DESIGN)

    # terminating the connection without providing resistance values results in a
    # warning
    app2.a.i2c.terminate(app2.a)
    check_design(app2.get_graph(), F.implements_design_check.CheckStage.POST_DESIGN)

    # setting a sufficient resistance fully satisfies the check
    app2.a.i2c.pull_up_sda.resistance.alias_is(0.2 * P.kohm)
    app2.a.i2c.pull_up_scl.resistance.alias_is(0.2 * P.kohm)
    check_design(app2.get_graph(), F.implements_design_check.CheckStage.POST_DESIGN)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.libs.exceptions import UserDesignCheckException, accumulate, downgrade

logger = logging.getLogger(__name__)


def check_design(
    G: Graph,
    stage: F.implements_design_check.CheckStage,
    exclude: tuple[str, ...] = tuple(),
):
    """
    args:
        exclude: list of names of checks to exclude e.g:
        - `I2C.requires_unique_addresses`
    """
    logger.info(f"Running design checks for stage {stage.name}")

    with accumulate(UserDesignCheckException) as accumulator:
        for _, trait in GraphFunctions(G).nodes_with_trait(F.implements_design_check):
            if trait.get_name_of_test() in exclude:
                continue

            with accumulator.collect():
                try:
                    trait.run(stage)
                except F.implements_design_check.MaybeUnfulfilledCheckException as e:
                    with downgrade(UserDesignCheckException):
                        raise UserDesignCheckException.from_nodes(
                            str(e), e.nodes
                        ) from e
                except F.implements_design_check.UnfulfilledCheckException as e:
                    raise UserDesignCheckException.from_nodes(str(e), e.nodes) from e



import os

import pytest

from faebryk.libs.util import get_module_from_path, hash_string, import_from_path


# Create test fixtures
@pytest.fixture
def temp_module(tmp_path):
    """Create a temporary Python module for testing"""
    module_path = tmp_path / "test_module.py"
    module_content = """
TEST_CONSTANT = "test_value"
def test_function():
    return "test_result"
"""
    module_path.write_text(module_content, encoding="utf-8")
    return module_path


@pytest.fixture
def temp_package(tmp_path):
    """Create a temporary Python package for testing"""
    package_dir = tmp_path / "test_package"
    package_dir.mkdir()

    init_path = package_dir / "__init__.py"
    init_path.write_text("PACKAGE_CONSTANT = 'package_value'", encoding="utf-8")

    submodule_path = package_dir / "submodule.py"
    submodule_path.write_text("SUB_CONSTANT = 'sub_value'", encoding="utf-8")

    return package_dir


@pytest.fixture
def temp_non_python_file(tmp_path):
    """Create a temporary non-Python file for testing"""
    non_python_path = tmp_path / "non_python_file.txt"
    non_python_path.write_text("This is a non-Python file", encoding="utf-8")
    return non_python_path


def test_hash_string():
    """Test that hash_string produces consistent results"""
    test_str = "test string"
    result1 = hash_string(test_str)
    result2 = hash_string(test_str)

    assert isinstance(result1, str)
    assert len(result1) > 0
    assert result1 == result2


def test_hash_string_different_inputs():
    """Test that hash_string produces different results for different inputs"""
    test_str1 = "test string 1"
    test_str2 = "test string 2"
    result1 = hash_string(test_str1)
    result2 = hash_string(test_str2)

    assert result1 != result2


def test_get_module_from_path_not_imported(temp_module):
    """Test get_module_from_path returns None for non-imported module"""
    result = get_module_from_path(temp_module)
    assert result is None


def test_get_module_from_path_imported(temp_module):
    """Test get_module_from_path returns module after import"""
    # First import the module
    module = import_from_path(temp_module)

    # Then try to get it
    result = get_module_from_path(temp_module)
    assert result is not None
    assert result == module
    assert hasattr(result, "TEST_CONSTANT")
    assert result.TEST_CONSTANT == "test_value"


def test_import_from_path_basic(temp_module):
    """Test basic module import functionality"""
    module = import_from_path(temp_module)

    assert module is not None
    assert hasattr(module, "TEST_CONSTANT")
    assert module.TEST_CONSTANT == "test_value"
    assert hasattr(module, "test_function")
    assert module.test_function() == "test_result"


def test_import_from_path_with_attr(temp_module):
    """Test importing specific attribute from module"""
    func = import_from_path(temp_module, "test_function")
    assert callable(func)
    assert func() == "test_result"


def test_import_from_path_nonexistent():
    """Test importing nonexistent module raises ImportError"""
    with pytest.raises(FileNotFoundError):
        import_from_path("nonexistent_module.py")


def test_import_from_path_not_python(temp_non_python_file):
    """Test importing non-Python file raises ImportError"""
    with pytest.raises(ImportError):
        import_from_path(temp_non_python_file)


def test_import_from_path_invalid_attr(temp_module):
    """Test importing nonexistent attribute raises AttributeError"""
    with pytest.raises(AttributeError):
        import_from_path(temp_module, "nonexistent_attr")


def test_import_from_path_reimport(temp_module):
    """Test that reimporting same module returns same instance"""
    module1 = import_from_path(temp_module)
    module2 = import_from_path(temp_module)

    assert module1 is module2
    assert id(module1) == id(module2)


def test_import_from_path_package(temp_package):
    """Test importing from a package"""
    module = import_from_path(temp_package / "__init__.py")

    assert module is not None
    assert hasattr(module, "PACKAGE_CONSTANT")
    assert module.PACKAGE_CONSTANT == "package_value"


def test_path_normalization(temp_module):
    """Test that different path formats resolve to same module"""
    # Get absolute path
    abs_path = temp_module.absolute()

    # Import using different path formats
    module1 = import_from_path(abs_path)
    module2 = import_from_path(str(abs_path))
    module3 = import_from_path(os.path.relpath(abs_path))

    assert module1 is module2 is module3



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.smd import SMDSize
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class CH344Q_ReferenceDesign(Module):
    """
    Minimal implementation of the CH344Q quad UART to USB bridge
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    usb: F.USB2_0
    usb_uart_converter: F.CH344Q
    oscillator: F.Crystal_Oscillator
    ldo: F.LDO
    led_rx = L.f_field(F.LEDIndicator)(use_mosfet=False, active_low=True)
    led_tx = L.f_field(F.LEDIndicator)(use_mosfet=False, active_low=True)
    led_act = L.f_field(F.LEDIndicator)(use_mosfet=False, active_low=True)
    power_led: F.PoweredLED
    reset_lowpass: F.FilterElectricalRC

    def __preinit__(self):
        # ------------------------------------
        #             aliases
        # ------------------------------------
        pwr_3v3 = self.usb_uart_converter.power
        # ------------------------------------
        #           connections
        # ------------------------------------
        self.usb_uart_converter.power.decoupled.decouple(owner=self).specialize(
            F.MultiCapacitor(4)
        ).set_equal_capacitance_each(L.Range.from_center_rel(100 * P.nF, 0.05))
        self.usb.usb_if.buspower.connect_via(self.ldo, pwr_3v3)

        self.usb.usb_if.d.connect(self.usb_uart_converter.usb)

        self.usb_uart_converter.act.connect(self.led_act.logic_in)
        self.usb_uart_converter.indicator_rx.connect(self.led_rx.logic_in)
        self.usb_uart_converter.indicator_tx.connect(self.led_tx.logic_in)
        pwr_3v3.connect(self.power_led.power)

        self.usb_uart_converter.osc[1].connect(self.oscillator.xtal_if.xin)
        self.usb_uart_converter.osc[0].connect(self.oscillator.xtal_if.xout)
        self.oscillator.xtal_if.gnd.connect(pwr_3v3.lv)

        self.reset_lowpass.out.line.connect(self.usb_uart_converter.reset.line)
        self.reset_lowpass.in_.line.connect(self.usb_uart_converter.reset.reference.hv)
        self.reset_lowpass.in_.reference.connect(
            self.usb_uart_converter.reset.reference
        )
        # TODO: already done by lowpass filter
        # self.usb_uart_converter.reset.pulled.pull(up=True)

        self.usb_uart_converter.test.set_weak(
            on=False, owner=self
        ).resistance.constrain_subset(L.Range.from_center_rel(4.7 * P.kohm, 0.05))

        self.ldo.enable_output()

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.usb_uart_converter.enable_status_or_modem_signals(self)

        self.oscillator.crystal.frequency.constrain_subset(
            L.Range.from_center_rel(8 * P.MHz, 0.001)
        )
        self.oscillator.crystal.frequency_tolerance.constrain_le(40 * P.ppm)
        self.oscillator.crystal.load_capacitance.constrain_subset(
            L.Range.from_center(8 * P.pF, 10 * P.pF)
        )  # TODO: should be property of crystal when picked
        self.oscillator.current_limiting_resistor.resistance.constrain_subset(0 * P.ohm)

        self.ldo.power_in.decoupled.decouple(owner=self).capacitance.constrain_subset(
            L.Range.from_center_rel(100 * P.nF, 0.1)
        )
        self.ldo.power_out.decoupled.decouple(owner=self).capacitance.constrain_subset(
            L.Range.from_center_rel(100 * P.nF, 0.1)
        )

        # self.usb.usb_if.buspower.max_current.constrain_subset(
        #    L.Range.from_center_rel(500 * P.mA, 0.1)
        # )

        self.ldo.output_current.constrain_ge(500 * P.mA)
        self.ldo.output_voltage.constrain_subset(
            L.Range.from_center_rel(3.3 * P.V, 0.05)
        )

        # reset lowpass
        # #TODO: fix
        # self.reset_lowpass.cutoff_frequency.constrain_subset(
        #    L.Range.from_center_rel(100 * P.Hz, 0.1)
        # )
        self.reset_lowpass.capacitor.capacitance.constrain_subset(
            L.Range.from_center_rel(100 * P.nF, 0.2)
        )
        self.reset_lowpass.resistor.resistance.constrain_subset(
            L.Range.from_center_rel(4.7 * P.kohm, 0.05)
        )

        for res in self.get_children(direct_only=True, types=F.Resistor):
            res.add(F.has_package_requirements(size=SMDSize.I0402))
        for cap in self.get_children(direct_only=True, types=F.Capacitor):
            cap.add(F.has_package_requirements(size=SMDSize.I0402))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import sys
from pathlib import Path

import platformdirs

# TODO change to use platformdirs for linux
# and  ~/.config/atopile for darwin


def get_config_dir() -> Path:
    if sys.platform == "win32":
        return Path(platformdirs.user_config_dir("atopile"))
    elif sys.platform == "linux":
        return Path.home() / "atopile"
    elif sys.platform == "darwin":
        return Path.home() / "atopile"
    else:
        return Path.home() / "atopile"


def get_data_dir() -> Path:
    if sys.platform == "win32":
        return Path(platformdirs.user_data_dir("atopile"))
    elif sys.platform == "linux":
        return Path.home() / "atopile"
    elif sys.platform == "darwin":
        return Path.home() / "atopile"
    else:
        return Path.home() / "atopile"


def get_cache_dir() -> Path:
    if sys.platform == "win32":
        return Path(platformdirs.user_cache_dir("atopile"))
    elif sys.platform == "linux":
        return Path.home() / "atopile" / "cache"
    elif sys.platform == "darwin":
        return Path.home() / "atopile" / "cache"
    else:
        return Path.home() / "atopile" / "cache"



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging
import re

import black
import black.parsing

logger = logging.getLogger(__name__)


def get_header():
    return (
        "# This file is part of the faebryk project\n# SPDX-License-Identifier: MIT\n"
    )


def formatted_file_contents(file_contents: str, is_pyi: bool = False) -> str:
    try:
        return black.format_str(
            file_contents,
            mode=black.Mode(
                is_pyi=is_pyi,
            ),
        )

    except black.parsing.InvalidInput as e:
        lineno, column = None, None
        match = re.match(r"Cannot parse: (\d+):(\d+):", e.args[0])
        if match:
            lineno, column = map(int, match.groups())
        with_line_numbers = "\n".join(
            f"{'>>' if i + 1 == lineno else '  '}{i + 1:3d}: {line}"
            for i, line in enumerate(file_contents.split("\n"))
        )
        logger.warning("black failed to format file:\n" + with_line_numbers)
        raise



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import pytest

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.core.solver.utils import Contradiction
from faebryk.libs.app.erc import ERCPowerSourcesShortedError, simple_erc
from faebryk.libs.library import L
from faebryk.libs.sets.quantity_sets import Quantity_Interval_Disjoint
from faebryk.libs.sets.sets import BoolSet
from faebryk.libs.units import P
from faebryk.libs.util import pairwise, times


@pytest.mark.xfail(reason="Solver not smart enough")
def test_fused_power():
    power_in = F.ElectricPower()
    power_out = F.ElectricPower()

    power_in.voltage.constrain_subset(10 * P.V)
    power_in.max_current.constrain_subset(500 * P.mA)

    power_in_fused = power_in.fused()
    power_in_fused.connect(power_out)

    fuse = next(iter(power_in_fused.get_children(direct_only=False, types=F.Fuse)))
    F.is_bus_parameter.resolve_bus_parameters(fuse.get_graph())

    solver = DefaultSolver()
    assert solver.inspect_get_known_supersets(fuse.trip_current).is_subset_of(
        L.Range.from_center_rel(500 * P.mA, 0.1)
    )
    assert solver.inspect_get_known_supersets(power_out.voltage).is_subset_of(
        L.Single(10 * P.V)
    )
    cur = solver.inspect_get_known_supersets(power_out.max_current)
    assert isinstance(cur, Quantity_Interval_Disjoint)
    assert (cur <= L.Single(500 * P.mA)) == BoolSet(True)


def test_power_source_short():
    """
    Test that a power source is shorted when connected to another power source
    """

    power_out_1 = F.ElectricPower()
    power_out_2 = F.ElectricPower()

    power_out_1.connect(power_out_2)
    power_out_2.connect(power_out_1)

    power_out_1.make_source()
    power_out_2.make_source()

    with pytest.raises(ERCPowerSourcesShortedError):
        simple_erc(power_out_1.get_graph())


def test_power_source_no_short():
    """
    Test that a power source is not shorted when connected to another non-power source
    """

    power_out_1 = F.ElectricPower()
    power_out_2 = F.ElectricPower()

    power_out_1.make_source()

    power_out_1.connect(power_out_2)

    simple_erc(power_out_1.get_graph())


def test_voltage_propagation():
    # Setup
    powers = times(4, F.ElectricPower)

    X = L.Range(10 * P.V, 15 * P.V)
    powers[0].voltage.constrain_subset(X)

    for p1, p2 in pairwise(powers):
        p1.connect(p2)

    F.is_bus_parameter.resolve_bus_parameters(powers[0].get_graph())

    # Test 1, propagate X from p[0] to p[-1]
    solver = DefaultSolver()
    solver.update_superset_cache(*powers)
    assert solver.inspect_get_known_supersets(powers[-1].voltage).is_subset_of(X)

    # Test 2, back propagate Y from p[-1] to p[0]
    Y = L.Single(10 * P.V)
    powers[-1].voltage.constrain_subset(Y)

    solver.update_superset_cache(*powers)
    y_back = solver.inspect_get_known_supersets(powers[0].voltage)
    assert y_back.is_subset_of(Y)


def test_current_consumption_sum_zero():
    class Test(Module):
        p1: F.ElectricPower
        p2: F.ElectricPower
        p3: F.ElectricPower

    test = Test()
    p1 = test.p1
    p2 = test.p2
    p3 = test.p3

    p1.connect(p2)
    p2.connect(p3)

    p1.max_current.alias_is(100 * P.mA)
    p2.max_current.alias_is(200 * P.mA)
    p3.max_current.alias_is(-300 * P.mA)

    p1.bus_max_current_consumption_sum.constrain_le(0 * P.mA)

    F.is_bus_parameter.resolve_bus_parameters(p1.get_graph())
    solver = DefaultSolver()
    solver.update_superset_cache(test)
    out = solver.inspect_get_known_supersets(p1.bus_max_current_consumption_sum)
    assert out.is_subset_of(L.Single(0 * P.mA))


def test_current_consumption_sum_negative():
    class Test(Module):
        p1: F.ElectricPower
        p2: F.ElectricPower
        p3: F.ElectricPower

    test = Test()
    p1 = test.p1
    p2 = test.p2
    p3 = test.p3

    p1.connect(p2)
    p2.connect(p3)

    p1.max_current.alias_is(300 * P.mA)
    p2.max_current.alias_is(200 * P.mA)
    p3.max_current.alias_is(-300 * P.mA)

    p1.bus_max_current_consumption_sum.constrain_le(0 * P.mA)

    F.is_bus_parameter.resolve_bus_parameters(p1.get_graph())
    solver = DefaultSolver()
    with pytest.raises(Contradiction):
        solver.update_superset_cache(test)
        solver.inspect_get_known_supersets(p1.bus_max_current_consumption_sum)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P
from faebryk.libs.util import times


class Header(Module):
    class PinType(Enum):
        MALE = auto()
        FEMALE = auto()

    class PadType(Enum):
        THROUGH_HOLE = auto()
        SMD = auto()

    class Angle(Enum):
        STRAIGHT = auto()
        ANGLE_90 = auto()

    def __init__(
        self,
        horizonal_pin_count: int,
        vertical_pin_count: int,
    ) -> None:
        super().__init__()
        self._horizontal_pin_count = horizonal_pin_count
        self._vertical_pin_count = vertical_pin_count

    def __preinit__(self):
        self.pin_count_horizonal.alias_is(self._horizontal_pin_count)
        self.pin_count_vertical.alias_is(self._vertical_pin_count)

    pin_pitch = L.p_field(
        units=P.mm,
        likely_constrained=True,
        domain=L.Domains.Numbers.REAL(),
        soft_set=L.Range(1 * P.mm, 10 * P.mm),
    )
    pin_type = L.p_field(
        domain=L.Domains.ENUM(PinType),
    )
    pad_type = L.p_field(
        domain=L.Domains.ENUM(PadType),
    )
    angle = L.p_field(
        domain=L.Domains.ENUM(Angle),
    )
    pin_count_horizonal = L.p_field(
        domain=L.Domains.Numbers.NATURAL(),
        soft_set=L.Range(2, 100),
    )
    pin_count_vertical = L.p_field(
        domain=L.Domains.Numbers.NATURAL(),
        soft_set=L.Range(2, 100),
    )

    mating_pin_length = L.p_field(
        units=P.mm,
        likely_constrained=True,
        domain=L.Domains.Numbers.REAL(),
        soft_set=L.Range(1 * P.mm, 10 * P.mm),
    )

    @L.rt_field
    def contact(self):
        return times(
            self._horizontal_pin_count * self._vertical_pin_count, F.Electrical
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )

    @L.rt_field
    def can_attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            pinmap={f"{i + 1}": self.contact[i] for i in range(len(self.contact))}
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class is_esphome_bus_defined(F.is_esphome_bus.impl()):
    def __init__(self, bus_id: str):
        super().__init__()
        self._bus_id = bus_id

    def get_bus_id(self) -> str:
        return self._bus_id



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from itertools import pairwise
from typing import cast

import pytest

import faebryk.library._F as F
from faebryk.core.node import Node
from faebryk.core.parameter import (
    Add,
    Additive,
    And,
    Expression,
    Is,
    Not,
    Parameter,
    ParameterOperatable,
)
from faebryk.libs.library import L
from faebryk.libs.library.L import Range
from faebryk.libs.sets.quantity_sets import Quantity_Interval, Quantity_Singleton
from faebryk.libs.sets.sets import BoolSet, EnumSet
from faebryk.libs.units import P
from faebryk.libs.util import times

logger = logging.getLogger(__name__)


def test_new_definitions():
    _ = Parameter(
        units=P.ohm,
        domain=L.Domains.Numbers.REAL(negative=False),
        soft_set=Range(1 * P.ohm, 10 * P.Mohm),
        likely_constrained=True,
    )


def test_compact_repr():
    p1 = Parameter(units=P.V)
    p2 = Parameter(units=P.V)
    context = ParameterOperatable.ReprContext()
    expr = cast(Expression, (p1 + p2 + (5 * P.V)) * 10)  # type: ignore
    exprstr = expr.compact_repr(context)
    assert exprstr == "((A volt + B volt) + 5 volt) * 10"
    expr2 = p2 + p1
    expr2str = expr2.compact_repr(context)
    assert expr2str == "B volt + A volt"

    p3 = Parameter(domain=L.Domains.BOOL())
    expr3 = Not(p3)
    expr3str = expr3.compact_repr(context)
    assert expr3str == "¬C"

    expr4 = And(expr3, (expr >= 10 * P.V))
    expr4str = expr4.compact_repr(context)
    assert expr4str == "¬C ∧ ((((A volt + B volt) + 5 volt) * 10) ≥ 10 volt)"

    manyps = times(ord("Z") - ord("C") - 1, Parameter)
    Additive.sum(manyps).compact_repr(context)

    pZ = Parameter()
    assert pZ.compact_repr(context) == "Z"

    pa = Parameter()
    assert pa.compact_repr(context) == "a"

    manyps2 = times(ord("z") - ord("a"), Parameter)
    Additive.sum(manyps2).compact_repr(context)
    palpha = Parameter()
    assert palpha.compact_repr(context) == "α"
    pbeta = Parameter()
    assert pbeta.compact_repr(context) == "β"

    manyps3 = times(ord("ω") - ord("β"), Parameter)
    Additive.sum(manyps3).compact_repr(context)
    pAA = Parameter()
    assert pAA.compact_repr(context) == "A₁"


def test_expression_congruence():
    p1 = Parameter()
    p2 = Parameter()
    p3 = Parameter()
    assert (p1 + p2).is_congruent_to(p1 + p2)
    assert (p1 + p2).is_congruent_to(p2 + p1)

    assert hash(Quantity_Singleton(0)) == hash(Quantity_Singleton(0))
    assert Quantity_Singleton(0) == Quantity_Singleton(0)
    assert Add(Quantity_Singleton(0), p2, p1).is_congruent_to(
        Add(p1, p2, Quantity_Singleton(0))
    )

    assert Add(Quantity_Interval(0, 1)).is_congruent_to(
        Add(Quantity_Interval(0, 1)), allow_uncorrelated=True
    )
    assert not (p1 - p2).is_congruent_to(p2 - p1)

    assert Is(p1, p2).is_congruent_to(Is(p2, p1))
    assert Is(p1, BoolSet(True)).is_congruent_to(Is(BoolSet(True), p1))
    p3.alias_is(p2)
    assert not Is(p1, p3).is_congruent_to(Is(p1, p2))


def test_expression_congruence_not():
    A = Parameter()
    x = Is(A, EnumSet(F.LED.Color.EMERALD))
    assert x.is_congruent_to(Is(A, EnumSet(F.LED.Color.EMERALD)))
    assert Not(x).is_congruent_to(Not(x))


@pytest.mark.skip
def test_visualize():
    """
    Creates webserver that opens automatically if run in jupyter notebook
    """
    from faebryk.exporters.visualize.interactive_params import visualize_parameters

    class App(Node):
        p1 = L.f_field(Parameter)(units=P.V)

    app = App()

    p2 = Parameter(units=P.V)
    p3 = Parameter(units=P.A)

    # app.p1.constrain_ge(p2 * 5)
    # app.p1.operation_is_ge(p2 * 5).constrain()
    (app.p1 >= p2 * 5).constrain()

    (p2 * p3 + app.p1 * 1 * P.A <= 10 * P.W).constrain()

    pytest.raises(ValueError, bool, app.p1 >= p2 * 5)

    G = app.get_graph()
    visualize_parameters(G, height=1400)


@pytest.mark.skip
def test_visualize_chain():
    from faebryk.exporters.visualize.interactive_params import visualize_parameters

    params = times(10, Parameter)
    sums = [p1 + p2 for p1, p2 in pairwise(params)]
    products = [p1 * p2 for p1, p2 in pairwise(sums)]
    bigsum = Additive.sum(products)

    predicates = [bigsum <= 100]
    for p in predicates:
        p.constrain()

    G = params[0].get_graph()
    visualize_parameters(G, height=1400)


@pytest.mark.skip
def test_visualize_inspect_app():
    from faebryk.exporters.visualize.interactive_params import visualize_parameters

    rp2040 = F.RP2040()

    G = rp2040.get_graph()
    visualize_parameters(G, height=1400)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_datasheet_defined(F.has_datasheet.impl()):
    def __init__(self, datasheet: str) -> None:
        super().__init__()
        self.datasheet = datasheet

    def get_datasheet(self) -> str:
        return self.datasheet
